__all__ = ('Invite',)

from datetime import datetime as DateTime, timedelta as TimeDelta

from scarletio import export

from ...bases import DiscordEntity
from ...core import INVITES
from ...http.urls import build_invite_url
from ...permission import Permission
from ...precreate_helpers import process_precreate_parameters_and_raise_extra
from ...user import ClientUserBase, ZEROUSER

from .fields import (
    parse_approximate_online_count, parse_approximate_user_count, parse_channel, parse_code, parse_created_at,
    parse_flags, parse_guild, parse_guild_activity_overview, parse_inviter, parse_max_age, parse_max_uses,
    parse_target_application, parse_target_type, parse_target_user, parse_temporary, parse_type, parse_user_permissions,
    parse_uses, put_approximate_online_count, put_approximate_user_count, put_channel, put_code, put_created_at,
    put_flags, put_guild, put_guild_activity_overview, put_inviter, put_max_age, put_max_uses, put_target_application,
    put_target_application_id, put_target_type, put_target_user, put_target_user_id, put_temporary, put_type,
    put_user_permissions, put_uses, validate_approximate_online_count, validate_approximate_user_count,
    validate_channel, validate_code, validate_created_at, validate_flags, validate_guild,
    validate_guild_activity_overview, validate_inviter, validate_max_age, validate_max_uses,
    validate_target_application, validate_target_type, validate_target_user, validate_temporary, validate_type,
    validate_user_permissions, validate_uses
)
from .flags import InviteFlag
from .preinstanced import InviteTargetType, InviteType


PRECREATE_FIELDS = {
    'approximate_user_count': ('approximate_user_count', validate_approximate_user_count),
    'approximate_online_count': ('approximate_online_count', validate_approximate_online_count),
    'channel': ('channel', validate_channel),
    'created_at': ('created_at', validate_created_at),
    'flags': ('flags', validate_flags),
    'guild': ('guild', validate_guild),
    'guild_activity_overview' : ('guild_activity_overview', validate_guild_activity_overview),
    'invite_type': ('type', validate_type),
    'inviter': ('inviter', validate_inviter),
    'max_age': ('max_age', validate_max_age),
    'max_uses': ('max_uses', validate_max_uses),
    'target_application': ('target_application', validate_target_application),
    'target_type': ('target_type', validate_target_type),
    'target_user': ('target_user', validate_target_user),
    'temporary': ('temporary', validate_temporary),
    'user_permissions': ('user_permissions', validate_user_permissions),
    'uses': ('uses', validate_uses),
}


@export
class Invite(DiscordEntity, immortal = True):
    """
    Represents a Discord Invite.
    
    Attributes
    ---------
    approximate_online_count : `int`
        The approximate amount of online users at the respective guild (or group channel).
        Defaults to `0`.
    
    approximate_user_count : `int`
        The approximate amount of users at the respective guild (or group channel).
        Defaults to `0`.
    
    channel : ``None | Channel``
        The channel where the invite redirects. If it is announcements or store channel, then the invite is a lurk
        invite. If channel data was not sent with the invite's, then this attribute is set as `None`.
    
    code : `str`
        The invite's unique identifier.
    
    created_at : `DateTime`
        When the invite was created. Defaults to Discord epoch.
    
    flags : ``InviteFlag``
        The invite's flags.
    
    guild : ``None | Guild``
        The guild the invite is for. If not included or if the invite's channel is a group channel, then set as
        `None`.
    
    guild_activity_overview : ``None | GuildActivityOverview``
        The guild's activity overview.
    
    inviter : ``ClientUserBase``
        The creator of the invite. If not included, then set as `ZEROUSER`.
    
    max_age : `None | int`
        The time in seconds after the invite will expire.
        Defaults to `None`.
        
        If the invite was created with max age as `0`, then this value will be negative instead of the expected `0`.
    
    max_uses : `None | int`
        How much times the invite can be used.
        Defaults to `None`.
        
        If the invite has no use limit, then this value is set as `0`.
    
    target_application : ``None | Application``
        The invite's target application.
    
    target_type : ``InviteTargetType``
        The invite's target type.
    
    target_user : ``None | ClientUserBase``
        The target of the invite if applicable.
    
    temporary : `bool`
        Whether this invite only grants temporary membership.
        
        When the user goes offline, they get kicked, except if they got a role meanwhile.
    
    type : ``InviteType``
        The invite's type.
    
    user_permissions : ``Permission``
        A subset of the permission that the user is granted by default upon joining.
    
    uses : `None | int`
        The amount how much times the invite was used.
        Defaults to `None`.
    """
    __slots__ = (
        'approximate_user_count', 'approximate_online_count', 'channel', 'code', 'created_at', 'flags', 'guild',
        'guild_activity_overview', 'inviter', 'max_age', 'max_uses', 'target_application', 'target_type',
        'target_user', 'temporary', 'type', 'user_permissions', 'uses'
    )
    
    
    def __new__(
        cls,
        *, 
        flags = ...,
        max_age = ...,
        max_uses = ...,
        target_application = ...,
        target_type = ...,
        target_user = ...,
        temporary = ...,
    ):
        """
        Creates a partial invite with the given fields.
        
        Parameters
        ----------
        flags : ``int | InviteFlag``, Optional (Keyword only)
            The invite's flags.
        
        max_age : `None | int`, Optional (Keyword only)
            The time in seconds after the invite will expire.
        
        max_uses : `None | int`, Optional (Keyword only)
            How much times the invite can be used.
        
        target_application : ``None | Application``, Optional (Keyword only)
            The invite's target application.
        
        target_type : ``None | int | InviteTargetType``, Optional (Keyword only)
            The invite's target type.
        
        target_user : ``None | ClientUserBase``, Optional (Keyword only)
            The target of the invite if applicable.
        
        temporary : `bool`, Optional (Keyword only)
            Whether this invite only grants temporary membership.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # flags
        if flags is ...:
            flags = InviteFlag()
        else:
            flags = validate_flags(flags)
        
        # max_age
        if max_age is ...:
            max_age = None
        else:
            max_age = validate_max_age(max_age)
        
        # max_uses
        if max_uses is ...:
            max_uses = None
        else:
            max_uses = validate_max_uses(max_uses)
        
        # target_application
        if target_application is ...:
            target_application = None
        else:
            target_application = validate_target_application(target_application)
        
        # target_type
        if target_type is ...:
            target_type = InviteTargetType.none
        else:
            target_type = validate_target_type(target_type)
        
        # target_user
        if target_user is ...:
            target_user = None
        else:
            target_user = validate_target_user(target_user)
        
        # temporary
        if temporary is ...:
            temporary = False
        else:
            temporary = validate_temporary(temporary)
        
        # Construct
        self = object.__new__(cls)
        
        self.approximate_online_count = 0
        self.approximate_user_count = 0
        self.channel = None
        self.code = ''
        self.created_at = None
        self.flags = flags
        self.guild = None
        self.guild_activity_overview = None
        self.inviter = ZEROUSER
        self.max_age = max_age
        self.max_uses = max_uses
        self.target_application = target_application
        self.target_type = target_type
        self.target_user = target_user
        self.temporary = temporary
        self.type = InviteType.guild
        self.user_permissions = Permission()
        self.uses = None
        
        return self
        
    
    @classmethod
    def from_data(cls, data):
        """
        Creates an invite from the given invite data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Invite data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        code = parse_code(data)
        
        try:
            self = INVITES[code]
        except KeyError:
            self = object.__new__(cls)
            self.code = code
            self._set_attributes(data)
            INVITES[code] = self
        
        else:
            self._update_attributes(data)
        
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the invite to json serializable dictionary.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default values should be included as well.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields, like id-s should be present as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        
        put_flags(self.flags, data, defaults)
        put_max_age(self.max_age, data, defaults)
        put_max_uses(self.max_uses, data, defaults)
        put_target_type(self.target_type, data, defaults)
        put_temporary(self.temporary, data, defaults)
        
        if include_internals:
            put_approximate_user_count(self.approximate_user_count, data, defaults)
            put_approximate_online_count(self.approximate_online_count, data, defaults)
            put_channel(self.channel, data, defaults)
            put_code(self.code, data, defaults)
            put_created_at(self.created_at, data, defaults)
            put_guild(self.guild, data, defaults)
            put_guild_activity_overview(self.guild_activity_overview, data, defaults)
            put_inviter(self.inviter, data, defaults)
            put_target_application(self.target_application, data, defaults)
            put_target_user(self.target_user, data, defaults)
            put_type(self.type, data, defaults)
            put_user_permissions(self.user_permissions, data, defaults)
            put_uses(self.uses, data, defaults)
        
        else:
            put_target_application_id(self.target_application_id, data, defaults)
            put_target_user_id(self.target_user_id, data, defaults)
        
        return data
    
    
    def __repr__(self):
        """Returns the representation of the invite."""
        return f'<{type(self).__name__} code = {self.code!r}>'
    
    
    def __hash__(self):
        """Returns the invite's code's hash."""
        code = self.code
        if code:
            return hash(code)
        return self._get_hash_partial()
    
    
    def _get_hash_partial(self):
        """
        Returns a partial invite's hash value.
        
        Returns
        -------
        hash_value : `int`
        """
        hash_value = 0
        
        # flags
        hash_value ^= self.flags
        
        # max_age
        hash_value ^= self.max_age << 4
        
        # max_uses
        hash_value ^= self.max_uses << 8
        
        # target_application
        target_application = self.target_application
        if (target_application is not None):
            hash_value ^= hash(target_application)
        
        # target_type
        hash_value ^= self.target_type.value << 12
        
        # target_user
        target_user = self.target_user
        if (target_user is not None):
            hash_value ^= hash(target_user)
        
        # temporary
        hash_value ^= self.temporary << 16
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two invites are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns whether the two invites are not equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether the two types are equal.
        
        Helper method for ``.__eq__``
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other instance. Must be from the same type.
        
        Returns
        -------
        is_equal : `bool`
        """
        # flags
        if self.flags != other.flags:
            return False
        
        # max_age
        if self.max_age != other.max_age:
            return False
        
        # max_uses
        if self.max_uses != other.max_uses:
            return False
        
        # target_application
        if self.target_application is not other.target_application:
            return False
        
        # target_type
        if self.target_type is not other.target_type:
            return False
        
        # target_user
        if self.target_user is not other.target_user:
            return False
        
        # temporary
        if self.temporary != other.temporary:
            return False
        
        return True
    
    
    def _set_attributes(self, data):
        """
        Sets the invite's attributes (except code).
        
        Parameters
        ----------
        data : `dict<str, object>`
            Invite data.
        """
        self._set_attributes_common(data)
        self._set_counts_only(data)
    
    
    def _update_attributes(self, data):
        """
        Updates the invite with the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Invite data.
        """
        self._set_attributes_common(data)
        self._update_counts_only(data)
    
    
    def _update_attributes_partial(self, data):
        """
        Updates the invite's fields that are received with a partial data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Invite data.
        """
        self.channel = parse_channel(data)
        self.guild = parse_guild(data)
    
    
    def _set_attributes_common(self, data):
        """
        Sets the common attributes of the invite. Used both in ``._set_attributes`` and ``._update_attributes``.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Invite data.
        """
        self.channel = parse_channel(data)
        self.created_at = parse_created_at(data)
        self.flags = parse_flags(data)
        self.guild = parse_guild(data)
        self.guild_activity_overview = parse_guild_activity_overview(data)
        self.inviter = parse_inviter(data)
        self.max_age = parse_max_age(data)
        self.max_uses = parse_max_uses(data)
        self.target_application = parse_target_application(data)
        self.target_type = parse_target_type(data)
        self.target_user = parse_target_user(data)
        self.temporary = parse_temporary(data)
        self.type = parse_type(data)
        self.user_permissions = parse_user_permissions(data)
        self.uses = parse_uses(data)
    
    
    def _set_counts_only(self, data):
        """
        Sets the invite's counts.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Received invite data.
        """
        guild = self.guild
        
        self.approximate_online_count = approximate_online_count = parse_approximate_online_count(data)
        if approximate_online_count and (guild is not None):
            guild.approximate_online_count = approximate_online_count
        
        self.approximate_user_count = approximate_user_count = parse_approximate_user_count(data)
        if approximate_user_count and (guild is not None):
            guild.approximate_user_count = approximate_user_count
    
    
    def _update_counts_only(self, data):
        """
        Updates the invite's counts if given.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Received invite data.
        """
        guild = self.guild
        
        approximate_online_count = parse_approximate_online_count(data)
        if approximate_online_count:
            self.approximate_online_count = approximate_online_count
            if (guild is not None):
                guild.approximate_online_count = approximate_online_count
        
        approximate_user_count = parse_approximate_user_count(data)
        if approximate_user_count:
            self.approximate_user_count = approximate_user_count
            if (guild is not None):
                guild.approximate_user_count = approximate_user_count
    
    
    @classmethod
    def precreate(cls, code, **keyword_parameters):
        """
        Precreates an invite object with the given parameters.
        
        Parameters
        ----------
        code : `str`
            The invite's code.
        **keyword_parameters : keyword parameters
            Additional predefined attributes for the invite.
        
        Other Parameters
        ----------------
        approximate_online_count : `int`, Optional (Keyword only)
            The amount of online users at the respective guild (or group channel).
        
        approximate_user_count : `int`, Optional (Keyword only)
            The amount of users at the respective guild (or group channel).
        
        channel : ``None | Channel``, Optional (Keyword only)
            The channel where the invite redirects.
        
        created_at : `None | DateTime`, Optional (Keyword only)
            When the invite was created.
        
        flags : ``int | InviteFlag``, Optional (Keyword only)
            The invite's flags.
        
        guild : ``None | Guild``, Optional (Keyword only)
            The guild the invite is for.
        
        guild_activity_overview : ``None | GuildActivityOverview``, Optional (Keyword only)
            The guild's activity overview.
        
        invite_type : ``None | int | InviteType``, Optional (Keyword only)
            The invite's type.
        
        inviter : ``ClientUserBase``, Optional (Keyword only)
            The creator of the invite.
        
        max_age : `None | int`, Optional (Keyword only)
            The time in seconds after the invite will expire.
        
        max_uses : `None | int`, Optional (Keyword only)
            How much times the invite can be used.
        
        target_application : ``None | Application``, Optional (Keyword only)
            The target application of the invite.
        
        target_type : ``int | InviteTargetType`` Optional (Keyword only)
            The invite's target type.
        
        target_user : ``None | ClientUserBase``, Optional (Keyword only)
            The target user of the invite.
        
        temporary : `bool`, Optional (Keyword only)
            Whether this invite only grants temporary membership.
        
        user_permissions : ``None | int | Permission``, Optional (Keyword only)
            A subset of the permission that the user is granted by default upon joining.
            
        uses : `None | int`, Optional (Keyword only)
            The amount how much times the invite was used.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        code = validate_code(code)

        if keyword_parameters:
            processed = process_precreate_parameters_and_raise_extra(keyword_parameters, PRECREATE_FIELDS)
        else:
            processed = None
        
        try:
            self = INVITES[code]
        except KeyError:
            self = cls._create_empty(code)
            INVITES[code] = self
        
        else:
            # Cannot detect if an invite is not partial -> do nothing
            pass
        
        if (processed is not None):
            for item in processed:
                setattr(self, *item)
        
        return self
    
    
    @classmethod
    def _create_empty(cls, code):
        """
        Creates an empty invite with default attributes set.
        
        Parameters
        ----------
        code : `str`
            Unique identifier of the invite.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        
        self.approximate_online_count = 0
        self.approximate_user_count = 0
        self.channel = None
        self.code = code
        self.created_at = None
        self.flags = InviteFlag()
        self.guild = None
        self.guild_activity_overview = None
        self.inviter = ZEROUSER
        self.max_age = None
        self.max_uses = None
        self.target_application = None
        self.target_type = InviteTargetType.none
        self.target_user = None
        self.temporary = False
        self.type = InviteType.guild
        self.user_permissions = Permission()
        self.uses = None
        
        return self
    
    
    def copy(self):
        """
        Copies the invite returning a new partial one.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        
        new.approximate_online_count = 0
        new.approximate_user_count = 0
        new.channel = None
        new.code = ''
        new.created_at = None
        new.flags = self.flags
        new.guild = None
        new.guild_activity_overview = None
        new.inviter = ZEROUSER
        new.max_age = self.max_age
        new.max_uses = self.max_uses
        new.target_application = self.target_application
        new.target_type = self.target_type
        new.target_user = self.target_user
        new.temporary = self.temporary
        new.type = InviteType.guild
        new.user_permissions = Permission()
        new.uses = None
        
        return new
    
    
    def copy_with(
        self,
        *, 
        flags = ...,
        max_age = ...,
        max_uses = ...,
        target_application = ...,
        target_type = ...,
        target_user = ...,
        temporary = ...,
    ):
        """
        Copies the invite with the given fields.
        
        Parameters
        ----------
        flags : ``int | InviteFlag``, Optional (Keyword only)
            The invite's flags.
        
        max_age : `None | int`, Optional (Keyword only)
            The time in seconds after the invite will expire.
        
        max_uses : `None | int`, Optional (Keyword only)
            How much times the invite can be used.
        
        target_application : ``None | Application``, Optional (Keyword only)
            The invite's target application.
        
        target_type : ``None | int | InviteTargetType``, Optional (Keyword only)
            The invite's target type.
        
        target_user : ``None | ClientUserBase``, Optional (Keyword only)
            The target of the invite if applicable.
        
        temporary : `bool`, Optional (Keyword only)
            Whether this invite only grants temporary membership.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # flags
        if flags is ...:
            flags = self.flags
        else:
            flags = validate_flags(flags)
        
        # max_age
        if max_age is ...:
            max_age = self.max_age
        else:
            max_age = validate_max_age(max_age)
        
        # max_uses
        if max_uses is ...:
            max_uses = self.max_uses
        else:
            max_uses = validate_max_uses(max_uses)
        
        # target_application
        if target_application is ...:
            target_application = self.target_application
        else:
            target_application = validate_target_application(target_application)
        
        # target_type
        if target_type is ...:
            target_type = self.target_type
        else:
            target_type = validate_target_type(target_type)
        
        # target_user
        if target_user is ...:
            target_user = self.target_user
        else:
            target_user = validate_target_user(target_user)
        
        # temporary
        if temporary is ...:
            temporary = self.temporary
        else:
            temporary = validate_temporary(temporary)
        
        # Construct
        new = object.__new__(type(self))
        
        new.approximate_online_count = 0
        new.approximate_user_count = 0
        new.channel = None
        new.code = ''
        new.created_at = None
        new.flags = flags
        new.guild = None
        new.guild_activity_overview = None
        new.inviter = ZEROUSER
        new.max_age = max_age
        new.max_uses = max_uses
        new.target_application = target_application
        new.target_type = target_type
        new.target_user = target_user
        new.temporary = temporary
        new.type = InviteType.guild
        new.user_permissions = Permission()
        new.uses = None
        
        return new
    
    
    @property
    def url(self):
        """
        Returns the invite's url.
        
        Returns
        -------
        url : `str`
        """
        return build_invite_url(self.code)
    
    
    @property
    def id(self):
        """
        Compatibility property with other Discord entities.
        
        Returns
        -------
        id : `int` = `0`
        """
        return 0
    
    
    @property
    def target_application_id(self):
        """
        Returns the invite's target application's identifier.
        
        Returns
        -------
        target_application_id : `int`
        """
        target_application = self.target_application
        if target_application is None:
            target_application_id = 0
        else:
            target_application_id = target_application.id
        
        return target_application_id
    
    
    @property
    def channel_id(self):
        """
        Returns the invite's channel's identifier.
        """
        channel = self.channel
        if channel is None:
            channel_id = 0
        else:
            channel_id = channel.id
        
        return channel_id
    
    
    @property
    def guild_id(self):
        """
        Returns the invite's guild's identifier.
        
        Returns
        -------
        guild_id : `int`
        """
        guild = self.guild
        if guild is None:
            guild_id = 0
        else:
            guild_id = guild.id
        
        return guild_id
    
    
    @property
    def inviter_id(self):
        """
        Returns the invite's creator's identifier.
        
        Returns
        -------
        user_id : `int`
        """
        return self.inviter.id
    
    
    @property
    def target_user_id(self):
        """
        Returns the invite's target user's identifier.
        
        Returns
        -------
        target_user_id : `int`
        """
        target_user = self.target_user
        if target_user is None:
            target_user_id = 0
        else:
            target_user_id = target_user.id
        
        return target_user_id
    
    
    @property
    def partial(self):
        """
        Returns whether the invite is partial.
        
        Since it is not possible to check whether the invite is up to date, it will always return `True` for
        non-template invites.
        
        Returns
        -------
        partial : `bool`
        """
        return False if self.code else True
    
    
    @property
    def expires_at(self):
        """
        Returns when the invite expires.
        
        Returns
        -------
        expires_at : `None | DateTime`
        """
        max_age = self.max_age
        if (max_age is not None):
            return self.created_at + TimeDelta(seconds = max_age)



__all__ = ()

from ...channel.channel.fields import validate_id
from ...channel.channel_metadata.fields import validate_status

from ..audit_log_entry_detail_conversion import AuditLogEntryDetailConversion, AuditLogEntryDetailConversionGroup
from ..conversion_helpers.converters import (
    value_deserializer_description, value_deserializer_id, value_serializer_description, value_serializer_id
)


# ---- id ----

ID_CONVERSION = AuditLogEntryDetailConversion(
    'channel_id',
    'id',
    value_deserializer = value_deserializer_id,
    value_serializer = value_serializer_id,
    value_validator = validate_id,
)


# ---- status ----

STATUS_CONVERSION = AuditLogEntryDetailConversion(
    'status',
    'status',
    value_deserializer = value_deserializer_description,
    value_serializer = value_serializer_description,
    value_validator = validate_status,
)


# ---- Construct ----

CHANNEL_CONVERSIONS = AuditLogEntryDetailConversionGroup(
    ID_CONVERSION,
    STATUS_CONVERSION,
)



__all__ = ()

from scarletio import export, include

from ...core import GUILDS

Guild = include('Guild')


@export
def _try_get_guild_id(guild):
    """
    Tries to get the guild's identifier.
    
    Parameters
    ----------
    guild : ``None | int | Guild``
        The guild or it's identifier.
    
    Returns
    -------
    guild_id : `int`
        The guild's identifier. Defaults to `0`.
    """
    if isinstance(guild, int):
        guild_id = guild
    elif guild is None:
        guild_id = 0
    elif isinstance(guild, Guild):
        guild_id = guild.id
    else:
        guild_id = 0
    
    return guild_id


def _try_get_guild_and_id(guild):
    """
    Tries to get the guild and it's identifier.
    
    Parameters
    ----------
    guild : ``None | int | Guild``
        The guild or it's identifier.
    
    Returns
    -------
    guild : ``None | Guild``
        The guild if found.
    guild_id : `int`
        The guild's identifier. Defaults to `0`.
    """
    if isinstance(guild, int):
        guild_id = guild
        guild = GUILDS.get(guild_id)
    elif guild is None:
        guild_id = 0
    elif isinstance(guild, Guild):
        guild_id = guild.id
    else:
        guild_id = 0
        guild = None
    
    return guild, guild_id



__all__ = ()


from ...field_parsers import (
    entity_id_array_parser_factory, entity_id_parser_factory, nullable_date_time_parser_factory,
    nullable_string_parser_factory, preinstanced_parser_factory
)
from ...field_putters import (
    entity_id_putter_factory, nullable_date_time_optional_putter_factory,
    optional_entity_id_array_optional_putter_factory, preinstanced_putter_factory, url_optional_putter_factory
)
from ...field_validators import (
    entity_id_array_validator_factory, entity_id_validator_factory, nullable_date_time_validator_factory,
    nullable_string_validator_factory, preinstanced_validator_factory
)
from ...user import ClientUserBase

from ..sku import SKU

from .preinstanced import SubscriptionStatus


# cancelled_at

parse_cancelled_at = nullable_date_time_parser_factory('cancelled_at')
put_cancelled_at = nullable_date_time_optional_putter_factory('cancelled_at')
validate_cancelled_at = nullable_date_time_validator_factory('cancelled_at')


# country_code

parse_country_code = nullable_string_parser_factory('country')
put_country_code = url_optional_putter_factory('country')
validate_country_code = nullable_string_validator_factory('country_code', 2, 2)


# current_period_end

parse_current_period_end = nullable_date_time_parser_factory('current_period_end')
put_current_period_end = nullable_date_time_optional_putter_factory('current_period_end')
validate_current_period_end = nullable_date_time_validator_factory('current_period_end')


# current_period_start

parse_current_period_start = nullable_date_time_parser_factory('current_period_start')
put_current_period_start = nullable_date_time_optional_putter_factory('current_period_start')
validate_current_period_start = nullable_date_time_validator_factory('current_period_start')


# entitlement_ids

parse_entitlement_ids = entity_id_array_parser_factory('entitlement_ids')
put_entitlement_ids = optional_entity_id_array_optional_putter_factory('entitlement_ids')
validate_entitlement_ids = entity_id_array_validator_factory('entitlement_ids', NotImplemented, include = 'Entitlement')


# id

parse_id = entity_id_parser_factory('id')
put_id = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('subscription_id')


# renewal_sku_ids

parse_renewal_sku_ids = entity_id_array_parser_factory('renewal_sku_ids')
put_renewal_sku_ids = optional_entity_id_array_optional_putter_factory('renewal_sku_ids')
validate_renewal_sku_ids = entity_id_array_validator_factory('renewal_sku_ids', SKU)


# sku_idpa

validate_sku_id = entity_id_validator_factory('sku_id', SKU)


# sku_ids

parse_sku_ids = entity_id_array_parser_factory('sku_ids')
put_sku_ids = optional_entity_id_array_optional_putter_factory('sku_ids')
validate_sku_ids = entity_id_array_validator_factory('sku_ids', SKU)


# status

parse_status = preinstanced_parser_factory('status', SubscriptionStatus, SubscriptionStatus.active)
put_status = preinstanced_putter_factory('status')
validate_status = preinstanced_validator_factory('status', SubscriptionStatus)


# user_id

parse_user_id = entity_id_parser_factory('user_id')
put_user_id = entity_id_putter_factory('user_id')
validate_user_id = entity_id_validator_factory('user_id', ClientUserBase)



from .event import *
from .event_deprecation import *
from .event_deprecation_descriptor import *
from .event_handler_plugin import *
from .helpers import *
from .meta import *


__all__ = (
    *event.__all__,
    *event_deprecation.__all__,
    *event_deprecation_descriptor.__all__,
    *event_handler_plugin.__all__,
    *helpers.__all__,
    *meta.__all__,
)



__all__ = ('MessageSearchResponse',)

from scarletio import RichAttributeErrorBaseType

from .fields import (
    parse_analytics_id, parse_deep_historical_indexing_in_progress, parse_messages, parse_result_count, parse_threads,
    put_analytics_id, put_deep_historical_indexing_in_progress, put_messages, put_result_count, put_threads,
    validate_analytics_id, validate_deep_historical_indexing_in_progress, validate_messages, validate_result_count,
    validate_threads
)


class MessageSearchResponse(RichAttributeErrorBaseType):
    """
    Response of a message search.
    
    Attributes
    ----------
    analytics_id : `int`
        Analytics identifier.
    
    deep_historical_indexing_in_progress : `bool`
        Whether deep historical indexing is in progress.
    
    messages : ``None | tuple<Messages>``
        The search result messages.
    
    result_count : `int`
        The total amount of messages matched.
    
    threads : ``None | tuple<Channel>``
        Additional threads to keep cached for channel resolution.
    """
    __slots__ = ('analytics_id', 'deep_historical_indexing_in_progress', 'messages', 'result_count', 'threads')
    
    def __new__(
        cls,
        *,
        analytics_id = ...,
        deep_historical_indexing_in_progress = ...,
        messages = ...,
        result_count = ...,
        threads = ...,
    ):
        """
        Creates a new message search response with the given fields.
        
        Parameters
        ----------
        analytics_id : `None | int`, Optional (Keyword only)
            Analytics identifier.
        
        deep_historical_indexing_in_progress : `None | bool`, Optional (Keyword only)
            Whether deep historical indexing is in progress.
        
        messages : ``None | iterable<Messages>``, Optional (Keyword only)
            The search result messages.
        
        result_count : `None | int`, Optional (Keyword only)
            The total amount of messages matched.
        
        threads : ``None | iterable<Channel>``, Optional (Keyword only)
            Additional threads to keep cached for channel resolution.
        
        Raises
        ------
        ValueError
            - If a parameter's value is incorrect.
        TypeError
            - If a parameter's type is incorrect.
        """
        # analytics_id
        if (analytics_id is ...):
            analytics_id = 0
        else:
            analytics_id = validate_analytics_id(analytics_id)
        
        # deep_historical_indexing_in_progress
        if (deep_historical_indexing_in_progress is ...):
            deep_historical_indexing_in_progress = False
        else:
            deep_historical_indexing_in_progress = validate_deep_historical_indexing_in_progress(
                deep_historical_indexing_in_progress,
            )
        
        # messages
        if (messages is ...):
            messages = None
        else:
            messages = validate_messages(messages)
        
        # result_count
        if (result_count is ...):
            result_count = 0
        else:
            result_count = validate_result_count(result_count)
        
        # threads
        if (threads is ...):
            threads = None
        else:
            threads = validate_threads(threads)
        
        # Construct
        self = object.__new__(cls)
        self.analytics_id = analytics_id
        self.deep_historical_indexing_in_progress = deep_historical_indexing_in_progress
        self.messages = messages
        self.result_count = result_count
        self.threads = threads
        return self
    
    
    def __repr__(self):
        """Returns repr(self)."""
        repr_parts = ['<', type(self).__name__]
        
        # analytics_id
        repr_parts.append(' analytics_id = ')
        repr_parts.append(repr(self.analytics_id))
        
        # deep_historical_indexing_in_progress
        repr_parts.append(', deep_historical_indexing_in_progress = ')
        repr_parts.append(repr(self.deep_historical_indexing_in_progress))
        
        # messages
        repr_parts.append(', messages = ')
        repr_parts.append(repr(self.messages))
        
        # result_count
        repr_parts.append(', result_count = ')
        repr_parts.append(repr(self.result_count))
        
        # threads
        repr_parts.append(', threads = ')
        repr_parts.append(repr(self.threads))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns hash(self)."""
        hash_value = 0
        
        # analytics_id
        analytics_id = self.analytics_id
        hash_value ^= (analytics_id & 0xffffffffffffffff) ^ (analytics_id >> 64)
        
        # deep_historical_indexing_in_progress
        hash_value ^= self.deep_historical_indexing_in_progress
        
        # messages
        messages = self.messages
        if (messages is not None):
            hash_value ^= len(messages) << 1
            
            for message in messages:
                hash_value ^= hash(message)
        
        # result_count
        hash_value ^= self.result_count << 6
        
        # threads
        threads = self.threads
        if (threads is not None):
            hash_value ^= len(threads) << 23
            
            for message in threads:
                hash_value ^= hash(message)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns self == other."""
        if type(self) is not type(other):
            return NotImplemented
        
        # analytics_id
        if self.analytics_id != other.analytics_id:
            return False
        
        # deep_historical_indexing_in_progress
        if self.deep_historical_indexing_in_progress != other.deep_historical_indexing_in_progress:
            return False
        
        # messages
        if self.messages != other.messages:
            return False
        
        # result_count
        if self.result_count != other.result_count:
            return False
        
        # threads
        if self.threads != other.threads:
            return False
        
        return True
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a message search response from the given data.
        
        Parameters
        ----------
        data : `dict<st, object>`
            Data to parse from.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.analytics_id = parse_analytics_id(data)
        self.deep_historical_indexing_in_progress = parse_deep_historical_indexing_in_progress(data)
        self.messages = parse_messages(data)
        self.result_count = parse_result_count(data)
        self.threads = parse_threads(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Serialises the message search response.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default values should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_analytics_id(self.analytics_id, data, defaults)
        put_deep_historical_indexing_in_progress(self.deep_historical_indexing_in_progress, data, defaults)
        messages = self.messages
        put_messages(messages, data, defaults)
        put_result_count(self.result_count, data, defaults)
        put_threads(self.threads, data, defaults, messages = messages)
        return data
    
    
    def copy(self):
        """
        Copies the message search response.
        
        Returns
        -------
        new : `instance<type<self>>˙
        """
        new = object.__new__(type(self))
        
        # analytics_id
        new.analytics_id = self.analytics_id
        
        # deep_historical_indexing_in_progress
        new.deep_historical_indexing_in_progress = self.deep_historical_indexing_in_progress
        
        # messages
        messages = self.messages
        if (messages is not None):
            messages = (*messages,)
        new.messages = messages
        
        # result_count
        new.result_count = self.result_count
        
        # threads
        threads = self.threads
        if (threads is not None):
            threads = (*threads,)
        new.threads = threads
        
        return new

    
    def copy_with(
        self,
        *,
        analytics_id = ...,
        deep_historical_indexing_in_progress = ...,
        messages = ...,
        result_count = ...,
        threads = ...,
    ):
        """
        Copies the message search response with the given fields..
        
        Parameters
        ----------
        analytics_id : `None | int`, Optional (Keyword only)
            Analytics identifier.
        
        deep_historical_indexing_in_progress : `None | bool`, Optional (Keyword only)
            Whether deep historical indexing is in progress.
        
        messages : ``None | iterable<Messages>``, Optional (Keyword only)
            The search result messages.
        
        result_count : `None | int`, Optional (Keyword only)
            The total amount of messages matched.
        
        threads : ``None | iterable<Channel>``, Optional (Keyword only)
            Additional threads to keep cached for channel resolution.
        
        Returns
        -------
        new : `instance<type<self>>˙
        
        Raises
        ------
        ValueError
            - If a parameter's value is incorrect.
        TypeError
            - If a parameter's type is incorrect.
        """
        # analytics_id
        if (analytics_id is ...):
            analytics_id = self.analytics_id
        else:
            analytics_id = validate_analytics_id(analytics_id)
        
        # deep_historical_indexing_in_progress
        if (deep_historical_indexing_in_progress is ...):
            deep_historical_indexing_in_progress = self.deep_historical_indexing_in_progress
        else:
            deep_historical_indexing_in_progress = validate_deep_historical_indexing_in_progress(
                deep_historical_indexing_in_progress
            )
        
        # messages
        if (messages is ...):
            messages = self.messages
            if (messages is not None):
                messages = (*messages,)
        else:
            messages = validate_messages(messages)
        
        # result_count
        if (result_count is ...):
            result_count = self.result_count
        else:
            result_count = validate_result_count(result_count)
        
        # threads
        if (threads is ...):
            threads = self.threads
            if (threads is not None):
                threads = (*threads,)
        else:
            threads = validate_threads(threads)
        
        # Construct
        new = object.__new__(type(self))
        new.analytics_id = analytics_id
        new.deep_historical_indexing_in_progress = deep_historical_indexing_in_progress
        new.messages = messages
        new.result_count = result_count
        new.threads = threads
        return new



__all__ = ('StatusByPlatform',)

from scarletio import RichAttributeErrorBaseType

from .fields import (
    parse_desktop, parse_embedded, parse_mobile, parse_web, put_desktop, put_embedded, put_mobile, put_web,
    validate_desktop, validate_embedded, validate_mobile, validate_platform, validate_web
)
from .preinstanced import SessionPlatformType, Status


class StatusByPlatform(RichAttributeErrorBaseType):
    """
    Represents an user's status by platform.
    
    Attributes
    ----------
    desktop : ``Status``
        The user's status on your average shitty electron application.
    
    embedded : ``Status``
        The user's status on an embedded platform.
    
    mobile : ``Status``
        The user's status on a mobile device.
    
    web : ``Status``
        The user's status in a web browser (excluding electron).
    """
    __slots__ = ('desktop', 'embedded', 'mobile', 'web')
    
    
    def __new__(cls, *, desktop = ..., embedded = ..., mobile = ..., web = ...):
        """
        Creates a new status by platform with the given parameters.
        
        Parameters
        ----------
        desktop : ``Status``, Optional (Keyword only)
            The user's status on your average shitty electron application.
        
        embedded : ``Status``
            The user's status on an embedded platform.
        
        mobile : ``Status``
            The user's status on a mobile device.
        
        web : ``Status``
            The user's status in a web browser (excluding electron).
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        """
        # desktop
        if desktop is ...:
            desktop = Status.offline
        else:
            desktop = validate_desktop(desktop)
        
        # embedded
        if embedded is ...:
            embedded = Status.offline
        else:
            embedded = validate_embedded(embedded)
        
        # mobile
        if mobile is ...:
            mobile = Status.offline
        else:
            mobile = validate_mobile(mobile)
        
        # web
        if web is ...:
            web = Status.offline
        else:
            web = validate_web(web)
        
        # Construct
        self = object.__new__(cls)
        self.desktop = desktop
        self.embedded = embedded
        self.mobile = mobile
        self.web = web
        return self
    
    
    def __repr__(self):
        """Returns repr(self)."""
        repr_parts = ['<', type(self).__name__]
        
        # desktop
        desktop = self.desktop
        if (desktop is Status.offline):
            field_added = False
        
        else:
            field_added = True
            
            repr_parts.append(' desktop = ')
            repr_parts.append(desktop.name)
        
        # embedded
        embedded = self.embedded
        if (embedded is not Status.offline):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' embedded = ')
            repr_parts.append(embedded.name)
        
        # mobile
        mobile = self.mobile
        if (mobile is not Status.offline):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' mobile = ')
            repr_parts.append(mobile.name)
        
        # web
        web = self.web
        if (web is not Status.offline):
            if field_added:
                repr_parts.append(',')
            
            repr_parts.append(' web = ')
            repr_parts.append(web.name)
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns hash(self)."""
        hash_value = 0
        
        # desktop
        hash_value ^= hash(self.desktop)
        
        # embedded
        hash_value ^= hash(self.embedded) << 8
        
        # mobile
        hash_value ^= hash(self.mobile) << 16
        
        # web
        hash_value ^= hash(self.web) << 24
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns self == other."""
        if type(self) is not type(other):
            return NotImplemented
        
        # desktop
        if self.desktop is not other.desktop:
            return False
        
        # embedded
        if self.embedded is not other.embedded:
            return False
        
        # mobile
        if self.mobile is not other.mobile:
            return False
        
        # web
        if self.web is not other.web:
            return False
        
        return True
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a status by platform from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Received status by platform data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.desktop = parse_desktop(data)
        self.embedded = parse_embedded(data)
        self.mobile = parse_mobile(data)
        self.web = parse_web(data)
        return self
    
    
    def to_data(self, *, defaults = False,):
        """
        Converts the status by platform to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        
        put_desktop(self.desktop, data, defaults)
        put_embedded(self.embedded, data, defaults)
        put_mobile(self.mobile, data, defaults)
        put_web(self.web, data, defaults)
        
        return data
    
    
    def copy(self):
        """
        Copies the status by platform.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.desktop = self.desktop
        new.embedded = self.embedded
        new.mobile = self.mobile
        new.web = self.web
        return new
    
    
    def copy_with(self, *, desktop = ..., embedded = ..., mobile = ..., web = ...):
        """
        Copies the status by platform and modifies the defined the defined fields of it.
        
        Parameters
        ----------
        desktop : ``Status``, Optional (Keyword only)
            The user's status on your average shitty electron application.
        
        embedded : ``Status``
            The user's status on an embedded platform.
        
        mobile : ``Status``
            The user's status on a mobile device.
        
        web : ``Status``
            The user's status in a web browser (excluding electron).
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
            - Extra or unused parameters.
        ValueError
            - If a parameter's value is incorrect.
        """
        # desktop
        if desktop is ...:
            desktop = self.desktop
        else:
            desktop = validate_desktop(desktop)
        
        # embedded
        if embedded is ...:
            embedded = self.embedded
        else:
            embedded = validate_embedded(embedded)
        
        # mobile
        if mobile is ...:
            mobile = self.mobile
        else:
            mobile = validate_mobile(mobile)
        
        # web
        if web is ...:
            web = self.web
        else:
            web = validate_web(web)
        
        # construct
        new = object.__new__(type(self))
        new.desktop = desktop
        new.embedded = embedded
        new.mobile = mobile
        new.web = web
        return new
    
    
    def __getitem__(self, platform):
        """
        Returns the status for the given platform.
        
        Parameters
        ----------
        platform : ``None | str | Platform``
            Platform to return status for.
        
        Returns
        -------
        status : ``Status``
        """
        platform = validate_platform(platform)
        
        try:
            slot = STATUS_BY_PLATFORM_SLOT_BY_PLATFORM[platform]
        except KeyError:
            status = Status.offline
        else:
            status = slot.__get__(self, type(self))
        
        return status
    
    
    def iter_status_by_platform(self):
        """
        Iterates over the platforms and their statuses.
        
        This function is an iterable generator.
        
        Yields
        ------
        item : ``(SessionPlatformType, Status)``
        """
        yield SessionPlatformType.desktop, self.desktop
        yield SessionPlatformType.embedded, self.embedded
        yield SessionPlatformType.mobile, self.mobile
        yield SessionPlatformType.web, self.web
    

STATUS_BY_PLATFORM_SLOT_BY_PLATFORM = {
    SessionPlatformType.desktop : StatusByPlatform.desktop,
    SessionPlatformType.embedded : StatusByPlatform.embedded,
    SessionPlatformType.mobile : StatusByPlatform.mobile,
    SessionPlatformType.web : StatusByPlatform.web,
}



__all__ = ('ChannelMetadataGuildVoiceBase',)

from scarletio import copy_docs

from .constants import BITRATE_DEFAULT, USER_LIMIT_DEFAULT
from .fields import (
    parse_bitrate, parse_region, parse_user_limit, parse_voice_engaged_since, put_bitrate, put_region, put_user_limit,
    put_voice_engaged_since, validate_bitrate, validate_region, validate_user_limit, validate_voice_engaged_since
)
from .preinstanced import VoiceRegion

from .guild_main_base import ChannelMetadataGuildMainBase


class ChannelMetadataGuildVoiceBase(ChannelMetadataGuildMainBase):
    """
    Guild voice channel metadata base.
    
    Attributes
    ----------
    _cache_permission : ``None | dict<int, Permission>``
        A `user_id` to ``Permission`` relation mapping for caching permissions. Defaults to `None`.
    
    bitrate : `int`
        The bitrate (in bits) of the voice channel.
    
    name : `str`
        The channel's name.
    
    parent_id : `int`
        The channel's parent's identifier.
    
    permission_overwrites : ``None | dict<int, PermissionOverwrite>``
        The channel's permission overwrites.
    
    position : `int`
        The channel's position.
    
    region : ``VoiceRegion``
        The voice region of the channel.
    
    user_limit : `int`
        The maximal amount of users, who can join the voice channel, or `0` if unlimited.
    
    voice_engaged_since : `None | DateTime`
        Since when the voice channel is engaged with.
    
    Class Attributes
    ----------------
    order_group: `int` = `2`
        The channel's order group used when sorting channels.
    """
    __slots__ = ('bitrate', 'region', 'user_limit', 'voice_engaged_since')
    
    order_group = 2
    
    
    def __new__(
        cls,
        *,
        bitrate = ...,
        name = ...,
        parent_id = ...,
        permission_overwrites = ...,
        position = ...,
        region = ...,
        user_limit = ...,
        voice_engaged_since = ...,
    ):
        """
        Creates a new guild voice base channel metadata from the given parameters.
        
        Parameters
        ----------
        bitrate : `int`, Optional (Keyword only)
            The bitrate (in bits) of the voice channel.
        
        name : `str`, Optional (Keyword only)
            The channel's name.
        
        parent_id : ``None | int | Channel``, Optional (Keyword only)
            The channel's parent's identifier.
        
        permission_overwrites : ``None | iterable<PermissionOverwrite>``, Optional (Keyword only)
            The channel's permission overwrites.
        
        position : `int`, Optional (Keyword only)
            The channel's position.
        
        region : ``None | str | VoiceRegion``, Optional (Keyword only)
            The voice region of the channel.
        
        user_limit : `int`, Optional (Keyword only)
            The maximal amount of users, who can join the voice channel, or `0` if unlimited.
        
        voice_engaged_since : `None | DateTime`, Optional (Keyword only)
            Since when the voice channel is engaged with.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # bitrate
        if bitrate is ...:
            bitrate = BITRATE_DEFAULT
        else:
            bitrate = validate_bitrate(bitrate)
        
        # region
        if region is ...:
            region = VoiceRegion.unknown
        else:
            region = validate_region(region)
        
        # user_limit
        if user_limit is ...:
            user_limit = 0
        else:
            user_limit = validate_user_limit(user_limit)
        
        # voice_engaged_since
        if voice_engaged_since is ...:
            voice_engaged_since = None
        else:
            voice_engaged_since = validate_voice_engaged_since(voice_engaged_since)
        
        # Construct
        self = ChannelMetadataGuildMainBase.__new__(
            cls,
            name = name,
            permission_overwrites = permission_overwrites,
            parent_id = parent_id,
            position = position,
        )
        self.bitrate = bitrate
        self.region = region
        self.user_limit = user_limit
        self.voice_engaged_since = voice_engaged_since
        return self
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildMainBase.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            bitrate = keyword_parameters.pop('bitrate', ...),
            name = keyword_parameters.pop('name', ...),
            parent_id = keyword_parameters.pop('parent_id', ...),
            permission_overwrites = keyword_parameters.pop('permission_overwrites', ...),
            position = keyword_parameters.pop('position', ...),
            region = keyword_parameters.pop('region', ...),
            user_limit = keyword_parameters.pop('user_limit', ...),
            voice_engaged_since = keyword_parameters.pop('voice_engaged_since', ...)
        )
    
    
    @copy_docs(ChannelMetadataGuildMainBase.__hash__)
    def __hash__(self):
        hash_value = ChannelMetadataGuildMainBase.__hash__(self)
        
        # bitrate
        hash_value ^= self.bitrate << 12
        
        # region
        hash_value ^= hash(self.region)
        
        # user_limit
        hash_value ^= self.user_limit << 15
        
        # voice_engaged_since
        voice_engaged_since = self.voice_engaged_since
        if (voice_engaged_since is not None):
            hash_value ^= hash(voice_engaged_since)
        
        return hash_value
    
    
    @copy_docs(ChannelMetadataGuildMainBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not ChannelMetadataGuildMainBase._is_equal_same_type(self, other):
            return False
        
        # bitrate
        if self.bitrate != other.bitrate:
            return False
        
        # region
        if self.region is not other.region:
            return False
        
        # user_limit
        if self.user_limit != other.user_limit:
            return False
        
        # voice_engaged_since
        if self.voice_engaged_since != other.voice_engaged_since:
            return False
        
        return True
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildMainBase._create_empty)
    def _create_empty(cls):
        self = super(ChannelMetadataGuildVoiceBase, cls)._create_empty()
        
        self.bitrate = BITRATE_DEFAULT
        self.region = VoiceRegion.unknown
        self.user_limit = USER_LIMIT_DEFAULT
        self.voice_engaged_since = None
        
        return self
    
    
    @copy_docs(ChannelMetadataGuildMainBase.copy)
    def copy(self):
        new = ChannelMetadataGuildMainBase.copy(self)
        new.bitrate = self.bitrate
        new.region = self.region
        new.user_limit = self.user_limit
        new.voice_engaged_since = self.voice_engaged_since
        return new
    
    
    def copy_with(
        self,
        *,
        bitrate = ...,
        name = ...,
        parent_id = ...,
        permission_overwrites = ...,
        position = ...,
        region = ...,
        user_limit = ...,
        voice_engaged_since = ...,
    ):
        """
        Copies the guild voice base channel metadata with the given fields.
        
        Parameters
        ----------
        bitrate : `int`, Optional (Keyword only)
            The bitrate (in bits) of the voice channel.
        
        name : `str`, Optional (Keyword only)
            The channel's name.
        
        parent_id : ``None | int | Channel``, Optional (Keyword only)
            The channel's parent's identifier.
        
        permission_overwrites : ``None | iterable<PermissionOverwrite>``, Optional (Keyword only)
            The channel's permission overwrites.
        
        position : `int`, Optional (Keyword only)
            The channel's position.
        
        region : ``None | str | VoiceRegion``, Optional (Keyword only)
            The voice region of the channel.
        
        user_limit : `int`, Optional (Keyword only)
            The maximal amount of users, who can join the voice channel, or `0` if unlimited.
        
        voice_engaged_since : `None | DateTime`, Optional (Keyword only)
            Since when the voice channel is engaged with.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # bitrate
        if bitrate is ...:
            bitrate = self.bitrate
        else:
            bitrate = validate_bitrate(bitrate)
        
        # region
        if region is ...:
            region = self.region
        else:
            region = validate_region(region)
        
        # user_limit
        if user_limit is ...:
            user_limit = self.user_limit
        else:
            user_limit = validate_user_limit(user_limit)
        
        # voice_engaged_since
        if voice_engaged_since is ...:
            voice_engaged_since = self.voice_engaged_since
        else:
            voice_engaged_since = validate_voice_engaged_since(voice_engaged_since)
        
        # Construct
        new = ChannelMetadataGuildMainBase.copy_with(
            self,
            name = name,
            permission_overwrites = permission_overwrites,
            parent_id = parent_id,
            position = position,
        )
        new.bitrate = bitrate
        new.region = region
        new.user_limit = user_limit
        new.voice_engaged_since = voice_engaged_since
        return new
    
    
    @copy_docs(ChannelMetadataGuildMainBase.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            bitrate = keyword_parameters.pop('bitrate', ...),
            name = keyword_parameters.pop('name', ...),
            parent_id = keyword_parameters.pop('parent_id', ...),
            permission_overwrites = keyword_parameters.pop('permission_overwrites', ...),
            position = keyword_parameters.pop('position', ...),
            region = keyword_parameters.pop('region', ...),
            user_limit = keyword_parameters.pop('user_limit', ...),
            voice_engaged_since = keyword_parameters.pop('voice_engaged_since', ...),
        )
    
    
    @copy_docs(ChannelMetadataGuildMainBase._update_attributes)
    def _update_attributes(self, data):
        ChannelMetadataGuildMainBase._update_attributes(self, data)
        
        # bitrate
        self.bitrate = parse_bitrate(data)
        
        # region
        self.region = parse_region(data)
        
        # user_limit
        self.user_limit = parse_user_limit(data)
    
    
    @copy_docs(ChannelMetadataGuildMainBase._difference_update_attributes)
    def _difference_update_attributes(self, data):
        old_attributes = ChannelMetadataGuildMainBase._difference_update_attributes(self, data)
        
        # bitrate
        bitrate = parse_bitrate(data)
        if self.bitrate != bitrate:
            old_attributes['bitrate'] = self.bitrate
            self.bitrate = bitrate
        
        # region
        region = parse_region(data)
        if self.region is not region:
            old_attributes['region'] = self.region
            self.region = region
        
        # user_limit
        user_limit = parse_user_limit(data)
        if self.user_limit != user_limit:
            old_attributes['user_limit'] = self.user_limit
            self.user_limit = user_limit
        
        return old_attributes
    
    
    @copy_docs(ChannelMetadataGuildMainBase._update_voice_engaged_since)
    def _update_voice_engaged_since(self, data):
        
        # voice_engaged_since
        self.voice_engaged_since = parse_voice_engaged_since(data)
    
    
    @copy_docs(ChannelMetadataGuildMainBase._difference_update_voice_engaged_since)
    def _difference_update_voice_engaged_since(self, data):
        old_attributes = {}

        # voice_engaged_since
        voice_engaged_since = parse_voice_engaged_since(data)
        if self.voice_engaged_since != voice_engaged_since:
            old_attributes['voice_engaged_since'] = self.voice_engaged_since
            self.voice_engaged_since = voice_engaged_since
        
        return old_attributes
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildMainBase.from_data)
    def from_data(cls, data):
        self = super(ChannelMetadataGuildVoiceBase, cls).from_data(data)
        
        # voice_engaged_since | Its only received with the initial channel payload.
        self.voice_engaged_since = parse_voice_engaged_since(data)
        
        return self
    
    
    @copy_docs(ChannelMetadataGuildMainBase.to_data)
    def to_data(self, *, defaults = False, include_internals = False):
        data = ChannelMetadataGuildMainBase.to_data(self, defaults = defaults, include_internals = include_internals)
        
        # bitrate
        put_bitrate(self.bitrate, data, defaults)
        
        # region
        put_region(self.region, data, defaults)
        
        # user_limit
        put_user_limit(self.user_limit, data, defaults)
        
        # voice_engaged_since
        put_voice_engaged_since(self.voice_engaged_since, data, defaults)
        
        return data



__all__ = ()

from scarletio import Compound, IgnoreCaseMultiValueDictionary
from scarletio.web_common.headers import AUTHORIZATION

from ...application import Application
from ...application_command import (
    ApplicationCommand, ApplicationCommandPermission, ApplicationCommandPermissionOverwrite,
    ApplicationCommandPermissionOverwriteTargetType
)
from ...application_command.application_command.constants import (
    LIMIT_GLOBAL as APPLICATION_COMMAND_LIMIT_GLOBAL, LIMIT_GUILD as APPLICATION_COMMAND_LIMIT_GUILD
)
from ...application_command.application_command_permission.constants import (
    PERMISSION_OVERWRITES_MAX as APPLICATION_COMMAND_PERMISSION_OVERWRITES_MAX
)
from ...http import DiscordApiClient
from ...oauth2 import Oauth2Access, Oauth2Scope, Oauth2User

from ..request_helpers import (
    get_application_command_and_id, get_application_command_id, get_application_command_id_nullable, get_guild_id
)


def _assert__application_id(application_id):
    """
    Asserts the the client's has `.application` synced by asserting whether it's value is not `0`.
    
    Parameters
    ----------
    application_id : `int`
        The client's application id.
    
    Raises
    ------
    AssertionError
        - If the client's application is not yet synced.
    """
    if application_id == 0:
        raise AssertionError(
            'The client\'s application is not yet synced.'
        )
    
    return True


def _assert__application_command(application_command):
    """
    Asserts whether the given ``application_command`` is indeed ``ApplicationCommand`` instance.
    
    Parameters
    ----------
    application_command : ``ApplicationCommand``
        The application command to checkout.
    
    Raises
    ------
    AssertionError
        - If `application_command` was not given as ``ApplicationCommand``.
    """
    if not isinstance(application_command, ApplicationCommand):
        raise AssertionError(
            f'`application_command` can be `{ApplicationCommand.__name__}`, got '
            f'{application_command.__class__.__name__}; {application_command!r}.'
        )
    
    return True


def _assert__new_application_command(new_application_command):
    """
    Asserts whether the given ``new_application_command`` is indeed ``ApplicationCommand`` instance.
    
    Parameters
    ----------
    new_application_command : ``ApplicationCommand``
        The application command to checkout.
    
    Raises
    ------
    AssertionError
        - If `new_application_command` was not given as ``ApplicationCommand``.
    """
    if not isinstance(new_application_command, ApplicationCommand):
        raise AssertionError(
            f'`new_application_command` can be `{ApplicationCommand.__name__}`, got '
            f'{new_application_command.__class__.__name__}; {new_application_command!r}.'
        )
    
    return True


def _assert__application_command_permission_edit__access_scope(access):
    """
    Asserts whether ``Client.application_command_permission_edit``'s `access` parameter has the required scopes.
    
    This function is only called when `access` is either ``Oauth2Access`` or ``Oauth2User`` instance.
    
    Parameters
    ----------
    access : ``Oauth2Access``, ``Oauth2User``
        A user's access token.
    
    Raises
    ------
    AssertionError
        - If the `access` lacks `applications.commands.permissions.update` scope.
    """
    if not access.has_scope(Oauth2Scope.applications_commands_permissions_update):
        raise AssertionError(
            f'The given `access` not grants `\'applications.commands.permissions.update\'` scope, '
            f'what is required, got {access!r}.'
        )
    
    return True


class ClientCompoundApplicationCommandEndpoints(Compound):
    
    application: Application
    api : DiscordApiClient
    
    async def application_command_global_get(self, application_command):
        """
        Requests the given global application command.
        
        This method is a coroutine.
        
        Parameters
        ----------
        application_command : ``ApplicationCommand``, `int`
            The application command, or it's id to request.
        
        Returns
        -------
        application_commands : ``ApplicationCommand``
            The received application command.
        
        Raises
        ------
        TypeError
            If `application_command` was not given neither as ``ApplicationCommand`` not as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        application_command, application_command_id = get_application_command_and_id(application_command)
        
        application_command_data = await self.api.application_command_global_get(
            application_id,
            application_command_id,
            {'with_localizations': True},
        )
        
        if application_command is None:
            application_command = ApplicationCommand.from_data(application_command_data)
        else:
            application_command._update_attributes(application_command_data)
        
        return application_command
    
    
    async def application_command_global_get_all(self):
        """
        Requests the client's global application commands.
        
        This method is a coroutine.
        
        Returns
        -------
        application_commands : `list` of ``ApplicationCommand``
            The received application commands.
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        data = await self.api.application_command_global_get_all(
            application_id,
            {'with_localizations': True},
        )
        
        return [ApplicationCommand.from_data(application_command_data) for application_command_data in data]
    
    
    async def application_command_global_create(self, application_command):
        """
        Creates a new global application command.
        
        > If there is an application command with the given name, will overwrite that instead.
        >
        > Each day only maximum only 200 global application command can be created.
        
        This method is a coroutine.
        
        Parameters
        ----------
        application_command : ``ApplicationCommand``
            The application command to create.
        
        Returns
        -------
        application_command : ``ApplicationCommand``
            The created application command.
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        The command will be available in all guilds after 1 hour.
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        assert _assert__application_command(application_command)
        
        data = application_command.to_data()
        data = await self.api.application_command_global_create(application_id, data)
        return ApplicationCommand.from_data(data)
    
    
    async def application_command_global_edit(self, old_application_command, new_application_command):
        """
        Edits a global application command.
        
        This method is a coroutine.
        
        Parameters
        ----------
        old_application_command : ``ApplicationCommand``, `int`
            The application command to edit. Can be given as the application command's id as well.
        new_application_command : ``ApplicationCommand``
            The application command to edit to.
        
        Returns
        -------
        application_command : ``ApplicationCommand``
            The edited application command.
        
        Raises
        ------
        TypeError
            If `old_application_command` was not given neither as ``ApplicationCommand`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        The updates will be available in all guilds after 1 hour.
        """
        old_application_command, application_command_id = get_application_command_and_id(old_application_command)
        
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        assert _assert__new_application_command(new_application_command)
        
        data = new_application_command.to_data()
        
        # Handle https://github.com/discord/discord-api-docs/issues/2525
        if (old_application_command is not None) and (old_application_command.name == data['name']):
            del data['name']
        
        await self.api.application_command_global_edit(application_id, application_command_id, data)
        return ApplicationCommand._from_edit_data(data, application_command_id, application_id)
    
    
    async def application_command_global_delete(self, application_command):
        """
        Deletes the given application command.
        
        Parameters
        ----------
        application_command : ``ApplicationCommand``, `int`
            The application command delete edit. Can be given as the application command's id as well.
        
        Raises
        ------
        TypeError
            If `application_command` was not given neither as ``ApplicationCommand`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        application_command_id = get_application_command_id(application_command)
        
        await self.api.application_command_global_delete(application_id, application_command_id)
    
    
    async def application_command_global_update_multiple(self, application_commands):
        """
        Takes an iterable of application commands, and updates the actual global ones.
        
        If a command exists with the given name, edits it, if not, will creates a new one.
        
        > The created application commands count to the daily limit.
        
        This method is a coroutine.
        
        Parameters
        ----------
        application_commands : `iterable` of ``ApplicationCommand``
            The application commands to update the existing ones with.
        
        Returns
        -------
        application_commands : `list` of ``ApplicationCommand``
            The edited and created application commands.
        
        Raises
        ------
        TypeError
            - If `application_commands` is not iterable.
            - If an element of `application_commands` is not ``ApplicationCommand`` instance.
        ValueError
            - If more than `100` ``ApplicationCommand`` is given.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        The commands will be available in all guilds after 1 hour.
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        
        if getattr(application_commands, '__iter__', None) is None:
            raise TypeError(
                f'`application_commands` can be an `iterable`, got '
                f'{application_commands.__class__.__name__}; {application_commands!r}.'
            )
        
        application_command_datas = []
        
        for application_command in application_commands:
            if not isinstance(application_command, ApplicationCommand):
                raise TypeError(
                    f'`application_commands` contains a not `{ApplicationCommand.__name__}` element, got: '
                    f'{application_command.__class__.__name__}; {application_command!r}; '
                    f'application_commands={application_commands!r}.'
                )
            
            application_command_datas.append(application_command.to_data())
        
        
        application_command_count = len(application_command_datas)
        # If we have no application commands to edit return~
        if application_command_count == 0:
            return []
        
        if application_command_count >= APPLICATION_COMMAND_LIMIT_GLOBAL:
            raise ValueError(
                f'Maximum {APPLICATION_COMMAND_LIMIT_GLOBAL} application command can be given, got '
                f'{application_command_count!r}; {application_commands!r}.'
            )
        
        application_command_datas = await self.api.application_command_global_update_multiple(
            application_id, application_command_datas
        )
        
        return [
            ApplicationCommand.from_data(application_command_data)
            for application_command_data in application_command_datas
        ]
    
    
    async def application_command_guild_get(self, guild, application_command):
        """
        Requests the given guild application command.
        
        Parameters
        ----------
        application_command : ``ApplicationCommand``, `int`
            The application command, or it's id to request.
        
        Returns
        -------
        application_commands : ``ApplicationCommand``
            The received application command.
        
        Raises
        ------
        TypeError
            - If `guild` was not given neither as``Guild`` nor `int`.
            - If `application_command` was not given neither as ``ApplicationCommand`` not as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        guild_id = get_guild_id(guild)
        
        application_command, application_command_id = get_application_command_and_id(application_command)
        
        application_command_data = await self.api.application_command_guild_get(
            application_id,
            guild_id,
            application_command_id,
            {'with_localizations': True},
        )
        
        if application_command is None:
            application_command = ApplicationCommand.from_data(application_command_data)
        else:
            application_command._update_attributes(application_command_data)
        
        return application_command
    
    
    async def application_command_guild_get_all(self, guild):
        """
        Requests the client's global application commands.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``int | Guild``
            The guild, which application commands will be requested.
        
        Returns
        -------
        application_commands : `list` of ``ApplicationCommand``
            The received application commands.
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as``Guild`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        guild_id = get_guild_id(guild)
        
        data = await self.api.application_command_guild_get_all(
            application_id,
            guild_id,
            {'with_localizations': True},
        )
        
        return [ApplicationCommand.from_data(application_command_data) for application_command_data in data]
    
    
    async def application_command_guild_create(self, guild, application_command):
        """
        Creates a new guild application command.
        
        > If there is an application command with the given name, will overwrite that instead.
        >
        > Each day only maximum only 200 guild application command can be created at each guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``int | Guild``
            The guild, where application commands will be created.
        application_command : ``ApplicationCommand``
            The application command to create.
        
        Returns
        -------
        application_command : ``ApplicationCommand``
            The created application command.
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as``Guild`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        assert _assert__application_command(application_command)
        
        guild_id = get_guild_id(guild)
        
        data = application_command.to_data()
        data = await self.api.application_command_guild_create(application_id, guild_id, data)
        return ApplicationCommand.from_data(data)
    
    
    async def application_command_guild_edit(self, guild, old_application_command, new_application_command):
        """
        Edits a guild application command.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``int | Guild``
            The guild, to what the application command is bound to.
        old_application_command : ``ApplicationCommand``, `int`
            The application command to edit. Can be given as the application command's id as well.
        new_application_command : ``ApplicationCommand``
            The application command to edit to.
        
        Raises
        ------
        TypeError
            - If `guild` was not given neither as``Guild`` nor `int`.
            - If `old_application_command` was not given neither as ``ApplicationCommand`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        old_application_command, application_command_id = get_application_command_and_id(old_application_command)
        
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        assert _assert__new_application_command(new_application_command)
        
        guild_id = get_guild_id(guild)
        data = new_application_command.to_data()
        
        # Handle https://github.com/discord/discord-api-docs/issues/2525
        if (old_application_command is not None) and (old_application_command.name == data['name']):
            del data['name']
        
        await self.api.application_command_guild_edit(application_id, guild_id, application_command_id, data)
    
    
    async def application_command_guild_delete(self, guild, application_command):
        """
        Deletes the given application command.
        
        Parameters
        ----------
        guild : ``int | Guild``
            The guild, to what the application command is bound to.
        application_command : ``ApplicationCommand``, `int`
            The application command delete edit. Can be given as the application command's id as well.
        
        Raises
        ------
        TypeError
            - If `guild` was not given neither as``Guild`` nor `int`.
            - If `application_command` was not given neither as ``ApplicationCommand`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        guild_id = get_guild_id(guild)
        
        application_command_id = get_application_command_id(application_command)
        
        await self.api.application_command_guild_delete(application_id, guild_id, application_command_id)
    
    
    async def application_command_guild_update_multiple(self, guild, application_commands):
        """
        Takes an iterable of application commands, and updates the guild's actual ones.
        
        If a command exists with the given name, edits it, if not, will creates a new one.
        
        > The created application commands count to the daily limit.
        
        This method is a coroutine.
        
        Parameters
        ----------
        application_commands : `iterable` of ``ApplicationCommand``
            The application commands to update the existing ones with.
        
        Returns
        -------
        application_commands : `list` of ``ApplicationCommand``
            The edited and created application commands.
        
        Raises
        ------
        TypeError
            - If `guild` was not given neither as``Guild`` nor `int`.
            - If `application_commands` is not iterable.
            - If an element of `application_commands` is not ``ApplicationCommand`` instance.
        ValueError
            - If more than `100` ``ApplicationCommand`` is given.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        guild_id = get_guild_id(guild)
        
        
        iterator = getattr(type(application_commands), '__iter__', None)
        if iterator is None:
            raise TypeError(
                f'`application_commands` can be an `iterable`, got '
                f'{application_commands.__class__.__name__}; {application_commands!r}.'
            )
        
        
        application_command_datas = []
        
        for application_command in iterator(application_commands):
            if not isinstance(application_command, ApplicationCommand):
                raise TypeError(
                    f'`application_commands` contains a not `{ApplicationCommand.__name__}` element, got: '
                    f'{application_command.__class__.__name__}; {application_command!r}; '
                    f'application_commands={application_commands!r}.'
                )
            
            application_command_datas.append(application_command.to_data())
        
        
        application_command_count = len(application_command_datas)
        # If we have no application commands to edit return~
        if application_command_count == 0:
            return []
        
        if application_command_count >= APPLICATION_COMMAND_LIMIT_GUILD:
            raise ValueError(
                f'Maximum {APPLICATION_COMMAND_LIMIT_GUILD} application command can be given, got '
                f'{application_command_count!r}; {application_commands!r}.'
            )
        
        application_command_datas = await self.api.application_command_guild_update_multiple(
            application_id, guild_id, application_command_datas
        )
        
        return [
            ApplicationCommand.from_data(application_command_data)
            for application_command_data in application_command_datas
        ]
    
    
    async def application_command_permission_get(self, guild, application_command):
        """
        Returns the permissions set for the given `application_command` in the given `guild`.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``int | Guild``
            The respective guild.
        application_command : ``ApplicationCommand``, `int`
            The respective application command.
        
        Returns
        -------
        permission : ``ApplicationCommandPermission``
            The requested permissions.
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as``Guild`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        If the application command has no permission overwrites in the guild, Discord will drop the following error:
        
        ```py
        DiscordException Not Found (404), code = 10066: Unknown application command permissions
        ```
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        guild_id = get_guild_id(guild)
        
        application_command_id = get_application_command_id(application_command)
        
        permission_data = await self.api.application_command_permission_get(
            application_id, guild_id, application_command_id
        )
        
        return ApplicationCommandPermission.from_data(permission_data)
    
    
    async def application_command_permission_edit(self, access, guild, application_command, permission_overwrites):
        """
        Edits the permissions of the given `application_command` in the given `guild`.
        
        > The new permissions will overwrite the existing permission of an application command.
        >
        > A command will lose it's permissions on rename.
        
        The endpoint requires oauth access with `applications.commands.permissions.update` scope.
        
        The user with the scope must have in the guild:
        
        - Permission to manage guild and roles.
        - Ability to invoke the respective command.
        - Permission to manage the resources that will be affected. The can be roles, users and channels depending on
            the permission types.
        
        This method is a coroutine.
        
        Parameters
        ----------
        access : ``Oauth2Access``, ``Oauth2User``, `str`
            A user's access token to use.
        
        guild : ``int | Guild``
            The respective guild.
        
        application_command : `None`, ``ApplicationCommand``, `int`
            The respective application command.
            
            Can be `None` (or `0`) if you want to edit the default overwrites for all guild level application commands.
            
        permission_overwrites : `None`, (`tuple`, `list` , `set`) of ``ApplicationCommandPermissionOverwrite``
            The new permission overwrites of the given application command inside of the guild.
            
            Give it as `None` to remove all existing one.
        
        Returns
        -------
        permissions : ``ApplicationCommandPermission``
            The application command's new permissions.
        
        Raises
        ------
        TypeError
            - If `guild` was not given neither as``Guild`` nor `int`.
            - If `permission_overwrites` was not given as `None`, `tuple`, `list`, `set`.
            - If `permission_overwrites` contains a non ``ApplicationCommandPermissionOverwrite`` element.
        ValueError
            - If `permission_overwrites` contains more than 100 elements.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        # application_id
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        # access_token
        if isinstance(access, (Oauth2Access, Oauth2User)):
            assert _assert__application_command_permission_edit__access_scope(access)
            access_token = access.access_token
        
        elif isinstance(access, str):
            access_token = access
        
        else:
            raise TypeError(
                f'`access` can be `{Oauth2Access.__name__}`, `{Oauth2User.__name__}` `str`'
                f', got {access.__class__.__name__}; {access!r}.'
            )
        
        
        guild_id = get_guild_id(guild)
        
        application_command_id = get_application_command_id_nullable(application_command)
        if application_command_id == 0:
            application_command_id = application_id
        
        permission_overwrite_datas = []
        if (permission_overwrites is not None):
            if getattr(permission_overwrites, '__iter__', None) is None:
                raise TypeError(
                    f'`permission_overwrites` can be `None`, `list`, `tuple` or '
                    f'`set`, got {permission_overwrites.__class__.__name__}; {permission_overwrites!r}.'
                )
            
            for permission_overwrite in permission_overwrites:
                if not isinstance(permission_overwrite, ApplicationCommandPermissionOverwrite):
                    raise TypeError(
                        f'`permission_overwrites` can contain `{ApplicationCommandPermissionOverwrite.__name__}` '
                        f'elements, got {permission_overwrite.__class__.__name__}; {permission_overwrite!r}; '
                        f'permission_overwrites={permission_overwrites!r}.'
                    )
                
                # We update channel permission overwrites with id of 0
                if permission_overwrite.target_type is ApplicationCommandPermissionOverwriteTargetType.channel:
                    permission_overwrite = permission_overwrite.copy_with(target = ('channel', guild_id - 1))
                
                elif permission_overwrite.target_type is ApplicationCommandPermissionOverwriteTargetType.role:
                    permission_overwrite = permission_overwrite.copy_with(target = ('role', guild_id))
                
                permission_overwrite_datas.append(permission_overwrite.to_data())
        
        
        permission_overwrite_datas_length = len(permission_overwrite_datas)
        if permission_overwrite_datas_length > APPLICATION_COMMAND_PERMISSION_OVERWRITES_MAX:
            raise ValueError(
                f'`permission_overwrites` can contain up to `{APPLICATION_COMMAND_PERMISSION_OVERWRITES_MAX}` '
                f'permission_overwrites, got {permission_overwrite_datas_length!r}; '
                f'{permission_overwrites!r}.'
            )
        
        data = {'permissions': permission_overwrite_datas}
        
        headers = IgnoreCaseMultiValueDictionary()
        headers[AUTHORIZATION] = f'Bearer {access_token}'
        
        permission_data = await self.api.application_command_permission_edit(
            application_id, guild_id, application_command_id, data, headers
        )
        
        return ApplicationCommandPermission.from_data(permission_data)
    
    
    async def application_command_permission_get_all_guild(self, guild):
        """
        Returns the permissions set for application commands in the given `guild`.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``int | Guild``
            The guild to request application command permissions from.
        
        Returns
        -------
        permission : `list` of ``ApplicationCommandPermission``
            The requested permissions for all the application commands in the guild.
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as``Guild`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        guild_id = get_guild_id(guild)
        
        permission_datas = await self.api.application_command_permission_get_all_guild(application_id, guild_id)
        
        return [ApplicationCommandPermission.from_data(permission_data) for permission_data in permission_datas]



__all__ = ()

from functools import partial as partial_func

from scarletio import RichAttributeErrorBaseType

from .conversion import ConversionMeta
from .descriptor import _conversion_descriptor_sort_key, ConversionDescriptor


def _request_type_attribute(type_name, base_types, type_attributes, attribute_name):
    """
    Allows requesting a constructed type's attribute before it was created.
    
    Parameters
    ----------
    type_name : `str`
        The to create type's name.
    base_types : `set<type>`
        Types to inherit from.
    type_attributes : `dict<str, object>`
            The type attributes of the type to be created.
    attribute_name : `str`
        The attribute's name.
    
    Returns
    -------
    attribute_value : `object`
    
    Raises
    ------
    RuntimeError
        - Missing attribute.
    """
    try:
        return type_attributes[attribute_name]
    except KeyError:
        pass
    
    for base_type in base_types:
        try:
            return getattr(base_type, attribute_name)
        except AttributeError:
            pass
    
    raise RuntimeError(
        f'{type_name!s} would not have requested attribute: {attribute_name!r}.'
    )


def _without_duplication(elements):
    """
    Returns a new list without duplication.
    
    Parameters
    ----------
    elements : `list`
        Elements to process.
    
    Returns
    -------
    output : `list`
    """
    output = []
    collected = set()
    old_length = 0
    
    for element in elements:
        collected.add(element)
        new_length = len(collected)
        if new_length == old_length:
            continue
        
        old_length = new_length
        output.append(element)
    
    return output


def _collect_meta_type_instances(meta_type, base_types):
    """
    Collects the types from the inherited ones.
    
    Parameters
    ----------
    meta_type : `type`
        Meta type to match.
    base_types : `tuple<type>`
        Types to filter from.
    
    Returns
    -------
    builder_types : `list<type>`
    """
    conversion_types = []
    
    for base_type in base_types:
        if isinstance(base_type, meta_type):
            conversion_types.append(base_type)

    return conversion_types


def _collect_assigned_conversions(builder_types, type_attributes):
    """
    Collects the assigned conversions.
    
    Parameters
    ----------
    builder_types : `list<type>`
        Inherited builder types to collect from.
    type_attributes : `dict<str, object>`
        The type attributes of the type to be created.
    
    Returns
    -------
    assigned_conversions : `dict<str, Conversion>`
    """
    assigned_conversions = {}
    for conversion_type in builder_types:
        # Remove anything that is already assigned
        to_remove = None
        for key in assigned_conversions.keys():
            if not hasattr(conversion_type, key):
                continue
            
            if to_remove is None:
                to_remove = []
            to_remove.append(key)
        
        if (to_remove is not None):
            for key in to_remove:
                del assigned_conversions[key]
        
        # Extend with new conversions 
        assigned_conversions.update(conversion_type.CONVERSIONS_ASSIGNED)
    
    # Remove from type attributes.
    to_remove = None
    for key in assigned_conversions.keys():
        if key not in type_attributes:
            continue
        
        if to_remove is None:
            to_remove = []
        to_remove.append(key)
    
    if (to_remove is not None):
        for key in to_remove:
            del assigned_conversions[key]
    
    # Extend with new conversions
    for attribute_name, attribute_value in type_attributes.items():
        if isinstance(type(attribute_value), ConversionMeta):
            assigned_conversions[attribute_name] = attribute_value
    
    return assigned_conversions


def _collect_default_conversions(builder_types, type_attributes):
    """
    Collects the default conversions.
    
    Parameters
    ----------
    builder_types : `list<type>`
        Inherited builder types to collect from.
    type_attributes : `dict<str, object>`
        The type attributes of the type to be created.
    
    Returns
    -------
    default_conversions : `list<Conversion>`
    """
    default_conversions = []
    
    for conversion_type in builder_types:
        added_conversions = getattr(conversion_type, 'CONVERSIONS_DEFAULT', None)
        if (added_conversions is not None):
            default_conversions.extend(added_conversions)
    
    added_conversions = type_attributes.pop('__conversions_default__', None)
    if (added_conversions is not None):
        default_conversions.extend(added_conversions)
    
    return _without_duplication(default_conversions)


def _create_conversion_descriptors(type_name, base_types, type_attributes, default_conversions, assigned_conversions):
    """
    Creates the conversion descriptors.
    
    Parameters
    ----------
    type_name : `str`
        The created type's name.
    base_types : `tuple<type>`
        The parent types.
    type_attributes : `dict<str, object>`
        The type attributes of the type to be created.
    default_conversions : `list<Conversion>`
        Default conversions to create descriptors for.
    assigned_conversions : `dict<str, Conversion>`
        Assigned descriptors to create descriptors for.
    
    Returns
    -------
    descriptors : `list<ConversionDescriptor>`
    """
    attribute_requester = partial_func(_request_type_attribute, type_name, base_types, type_attributes)
    descriptors = []
    
    for conversion in default_conversions:
        descriptor = ConversionDescriptor(None, conversion, attribute_requester)
        descriptors.append(descriptor)
    
    for attribute_name, conversion in assigned_conversions.items():
        descriptor = ConversionDescriptor(attribute_name, conversion, attribute_requester)
        type_attributes[attribute_name] = descriptor
        descriptors.append(descriptor)
    
    return descriptors


def _filter_descriptors_typed(conversion_descriptors):
    """
    Creates a type to descriptor relation for fast typed validation.
    
    Parameters
    ----------
    conversion_descriptors : `list<ConversionDescriptor>`
        Descriptors to filter from.
    
    Returns
    -------
    descriptors_typed : `dict<type, ConversionDescriptor>`
    descriptors_typed_ordered : `list<ConversionDescriptor>`
    """
    descriptors_typed_ordered = []
    for conversion_descriptor in conversion_descriptors:
        if conversion_descriptor.conversion.set_type is not None:
            descriptors_typed_ordered.append(conversion_descriptor)
    
    descriptors_typed_ordered.sort(key = _conversion_descriptor_sort_key)
    descriptors_typed = {
        conversion_descriptor.conversion.set_type: conversion_descriptor
        for conversion_descriptor in descriptors_typed_ordered
    }
    return descriptors_typed, descriptors_typed_ordered


def _filter_descriptors_positional(conversion_descriptors):
    """
    Creates the descriptor listing for positional parameter input validation.
    
    Parameters
    ----------
    conversion_descriptors : `list<ConversionDescriptor>`
        Descriptors to filter from.
    
    Returns
    -------
    descriptors_positional : `list<ConversionDescriptor>`
    """
    descriptors_positional = []
    for conversion_descriptor in conversion_descriptors:
        if conversion_descriptor.conversion.set_identifier is not None:
            descriptors_positional.append(conversion_descriptor)
    
    descriptors_positional.sort(key = _conversion_descriptor_sort_key)
    return descriptors_positional


def _filter_descriptors_listing(conversion_descriptors):
    """
    Creates the descriptor listing for listing parameter input validation.
    
    Parameters
    ----------
    conversion_descriptors : `list<ConversionDescriptor>`
        Descriptors to filter from.
    
    Returns
    -------
    descriptors_listing : `list<ConversionDescriptor>`
    """
    descriptors_listing = []
    for conversion_descriptor in conversion_descriptors:
        if conversion_descriptor.conversion.set_listing_identifier is not None:
            descriptors_listing.append(conversion_descriptor)
    
    descriptors_listing.sort(key = _conversion_descriptor_sort_key)
    return descriptors_listing


def _filter_descriptors_keyword(conversion_descriptors):
    """
    Creates a descriptor mapping for keyword parameter input validation.
    
    Parameters
    ----------
    conversion_descriptors : `list<ConversionDescriptor>`
        Descriptors to filter from.
    
    Returns
    -------
    descriptors_keyword : `dict<str, ConversionDescriptor>`
    """
    descriptors_keyword = {}
    for conversion_descriptor in conversion_descriptors:
        if conversion_descriptor.conversion.set_validator is not None:
            for name in conversion_descriptor.conversion.iter_names():
                descriptors_keyword[name] = conversion_descriptor
    
    return descriptors_keyword


class BuilderMeta(type):
    """
    Builder metatype.
    """
    def __new__(cls, type_name, base_types, type_attributes):
        """
        Creates a new builder type.
        
        Parameters
        ----------
        type_name : `str`
            The created type's name.
        base_types : `tuple<type>`
            The parent types.
        type_attributes : `dict<str, object>`
            The type attributes of the type to be created.
        """
        builder_types = _collect_meta_type_instances(cls, base_types)
        
        # Collect CONVERSIONS_DEFAULT
        default_conversions = _collect_default_conversions(builder_types, type_attributes)
        
        # Collect CONVERSIONS_ASSIGNED
        assigned_conversions = _collect_assigned_conversions(builder_types, type_attributes)
        
        # create conversion_descriptors
        conversion_descriptors = _create_conversion_descriptors(
            type_name, base_types, type_attributes, default_conversions, assigned_conversions
        )
        
        # filter DESCRIPTORS_POSITIONAL
        descriptors_positional = _filter_descriptors_positional(conversion_descriptors)
        
        # filter DESCRIPTORS_KEYWORD
        descriptors_keyword = _filter_descriptors_keyword(conversion_descriptors)
        
        # filter DESCRIPTORS_TYPED, DESCRIPTORS_TYPED_ORDERED
        descriptors_typed, descriptors_typed_ordered = _filter_descriptors_typed(conversion_descriptors)
        
        # filter DESCRIPTORS_LISTING
        descriptors_listing = _filter_descriptors_listing(conversion_descriptors)
        
        # set
        type_attributes['CONVERSIONS_ASSIGNED'] = assigned_conversions
        type_attributes['CONVERSIONS_DEFAULT'] = default_conversions
        type_attributes['DESCRIPTORS_POSITIONAL'] = descriptors_positional
        type_attributes['DESCRIPTORS_KEYWORD'] = descriptors_keyword
        type_attributes['DESCRIPTORS_TYPED'] = descriptors_typed
        type_attributes['DESCRIPTORS_TYPED_ORDERED'] = descriptors_typed_ordered
        type_attributes['DESCRIPTORS_LISTING'] = descriptors_listing
        
        # create type
        
        return type.__new__(cls, type_name, base_types, type_attributes)


class BuilderBase(RichAttributeErrorBaseType, metaclass = BuilderMeta):
    """
    Base builder defining default functionality.
    """
    __slots__ = ()
    
    __conversions_default__ = None
    
    def __new__(cls):
        """
        Creates a new message builder.
        """
        self = object.__new__(cls)
        return self
    
    
    def __eq__(self, other):
        """Returns whether the two builders are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        for conversion in {*self._iter_conversions(), *other._iter_conversions()}:
            if self._getter_field(conversion) != other._getter_field(conversion):
                return False
        
        return True
    
    
    def __hash__(self):
        """Returns the hash value of the builder."""
        hash_value = 0
        
        for conversion, value in self._iter_fields():
            if not value:
                continue
            
            hash_value ^= hash(conversion) & hash(value)
        
        return hash_value
    
    
    def __repr__(self):
        """Returns the builder's representation."""
        repr_parts = ['<', type(self).__name__]
        
        field_added = False
        
        for conversion, value in self._iter_fields():
            if not value:
                continue
            
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' ')
            repr_parts.append(conversion.name)
            repr_parts.append(' = ')
            repr_parts.append(repr(value))
            
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def _setter_none(self, conversion, value):
        """
        Sets nothing.
        
        Parameters
        ----------
        conversion : ``Conversion``
            Conversion we are setting.
        value : `object`
            The value to set.
        
        Returns
        -------
        field_set : `bool`
        """
        return True
    
    
    def _setter_field(self, conversion, value):
        """
        Sets the given field value.
        
        Parameters
        ----------
        conversion : ``Conversion``
            Conversion we are setting.
        value : `object`
            The value to set.
        
        Returns
        -------
        field_set : `bool`
        """
        set_merger = conversion.set_merger
        if (set_merger is not None):
            for old_value in self._try_pull_field_value(conversion):
                value = set_merger(old_value, value)
        
        self._store_field_value(conversion, value)
        return True
    
    
    def _store_field_value(self, conversion, value):
        """
        Stores the given field value.
        
        This method is method is an iterable generator.
        
        Parameters
        ----------
        conversion : ``Conversion``
            Conversion we are setting.
        value : `object`
            The value to set.    
        """
        raise NotImplemented
    
    
    def _try_pull_field_value(self, conversion):
        """
        Tries to get the field of the given conversion.
        
        This method is an iterable generator that should yield the matched `field` value once and not yield anything if
        not.
        
        Yields
        ------
        value : `object`
        """
        raise NotImplemented
        yield
    
    
    def _setter_positional(self, conversion, value):
        """
        Sets the given positional field values.
        
        Parameters
        ----------
        conversion : ``Conversion``
            Conversion we are setting.
        value : `object`
            The value to set.
        
        Returns
        -------
        field_set : `bool`
        """
        self._with_positional_parameters(value)
        return True
    
    
    def _setter_keyword(self, conversion, value):
        """
        Sets the given keyword field values.
        
        Parameters
        ----------
        conversion : ``Conversion``
            Conversion we are setting.
        value : `object`
            The value to set.
        
        Returns
        -------
        field_set : `bool`
        """
        self._with_keyword_parameters(value)
        return True
    
    
    def _setter_instance(self, conversion, value):
        """
        Sets the given builder instance's fields into self.
        
        Parameters
        ----------
        conversion : ``Conversion``
            Conversion we are setting.
        value : `instance<type<self>>`
            The value to set.
        
        Returns
        -------
        field_set : `bool`
        """
        for field_conversion, field_value in value._iter_fields():
            self._setter_field(field_conversion, field_value)
        return True
    
    
    def _with_positional_parameter_unknown(self, value):
        """
        Handles an unknown positional parameter adding. Called by ``._with_positional_parameters``.
        
        Parameters
        ----------
        value : `object`
            The value to add.
        
        Returns
        -------
        added : `bool`
        """
        return True
    
    
    def _try_match_as_typed(self, value):
        """
        Tries to match the given value based on type and then sets it.
        
        Parameters
        ----------
        value : `object`
            Value to match.
        
        Returns
        -------
        matched : `bool`
        """
        try:
            descriptor = self.DESCRIPTORS_TYPED[type(value)]
        except KeyError:
            return False
        
        set_type_processor = descriptor.conversion.set_type_processor
        if set_type_processor is None:
            processed_value = value
        else:
            processed_value = set_type_processor(value)
        
        descriptor.setter(self, descriptor.output_conversion, processed_value)
        return True
    
    
    def _try_match_as_listing(self, value):
        """
        Tries to match the given value if its a listing and then sets it.
        
        Parameters
        ----------
        value : `object`
            Value to match.
        
        Returns
        -------
        matched : `bool`
        """
        if not isinstance(value, list) or not value:
            return False
        
        for descriptor in self.DESCRIPTORS_LISTING:
            for processed_value in descriptor.conversion.set_listing_identifier(value):
                descriptor.setter(self, descriptor.output_conversion, processed_value)
                return True
        
        return False
    
    
    def _try_match_as_sub_type(self, value):
        """
        Tries to match the given value based on type and then sets it.
        Checks for inheritance and if matched updates the match table.
        
        Parameters
        ----------
        value : `object`
            Value to match.
        
        Returns
        -------
        matched : `bool`
        """
        value_type = type(value)
        for descriptor in self.DESCRIPTORS_TYPED_ORDERED:
            if issubclass(value_type, descriptor.conversion.set_type):
                self.DESCRIPTORS_TYPED[value_type] = descriptor
                break
        else:
            return False
                
        set_type_processor = descriptor.conversion.set_type_processor
        if set_type_processor is None:
            processed_value = value
        else:
            processed_value = set_type_processor(value)
        
        descriptor.setter(self, descriptor.output_conversion, processed_value)
        return True
    
    
    def _try_match_as_identified(self, value):
        """
        Tries to match the given value based on identifiers and then sets it.
        
        Parameters
        ----------
        value : `object`
            Value to match.
        
        Returns
        -------
        matched : `bool`
        """
        for descriptor in self.DESCRIPTORS_POSITIONAL:
            for processed_value in descriptor.conversion.set_identifier(value):
                descriptor.setter(self, descriptor.output_conversion, processed_value)
                return True
        
        return False
    
    
    def _with_positional_parameters(self, positional_parameters):
        """
        Extends the builder with the given positional parameters.
        
        Parameters
        ----------
        positional_parameters : `iterable<object>`
            Positional parameters to extend self with.
        
        Raises
        ------
        TypeError
            - Value of invalid type given.
            - Unrecognised value.
        """
        for value in positional_parameters:
            if self._try_match_as_typed(value):
                continue
            
            if self._try_match_as_listing(value):
                continue
            
            if self._try_match_as_sub_type(value):
                continue
            
            if self._try_match_as_identified(value):
                continue
            
            self._with_positional_parameter_unknown(value)
    
    
    def _with_keyword_parameter_unknown(self, key, value):
        """
        Handles an unknown keyword parameter adding. Called by ``._with_keyword_parameters``.
        
        Parameters
        ----------
        key : `str`
            They key to add for.
        
        value : `object`
            The value to add.
        
        Returns
        -------
        added : `bool`
        """
        return True
    
    
    def _with_keyword_parameters(self, keyword_parameters):
        """
        Extends the builder with the given keyword parameters.
        
        Parameters
        ----------
        keyword_parameters : `dict<str, object>`
            Keyword parameters to extend self with.
        
        Raises
        ------
        TypeError
            - Value of invalid type given.
            - Unrecognised value.
        """
        descriptors_keyword = self.DESCRIPTORS_KEYWORD
        
        for key, value in keyword_parameters.items():
            try:
                descriptor = descriptors_keyword[key]
            except KeyError:
                pass
            else:
                for processed_value in descriptor.conversion.set_validator(value):
                    descriptor.setter(self, descriptor.output_conversion, processed_value)
                    break
                
                else:
                    return descriptor.raise_type_error(key, value)
                
                continue
            
            self._with_keyword_parameter_unknown(key, value)
            continue
    
    
    def _getter_none(self, conversion):
        """
        Default value getter. Rejects getting.
        
        Parameters
        ----------
        conversion : ``Conversion``
            Conversion we are getting.
        
        Returns
        -------
        value : `object`
        
        Raises
        ------
        RuntimeError
            - Field is not gettable.
        """
        raise RuntimeError(
            f'Field {conversion.name} is not gettable.'
        )
    
    
    def _getter_field(self, conversion):
        """
        Gets the field value for the given conversion.
        
        Parameters
        ----------
        conversion : ``Conversion``
            Conversion to pull the value for.
        
        Returns
        -------
        field_value : `object`
        """
        for value in self._try_pull_field_value(conversion):
            break
        else:
            return conversion.get_default
        
        get_processor = conversion.get_processor
        if (get_processor is not None):
            value = get_processor(value)
        
        return value
    
    
    def _iter_conversions(self):
        """
        Iterates over the set conversions of the builder.
        
        This method is an iterable generator.
        
        Yields
        ------
        conversion : ``Conversion``
        """
        return
        yield
    
    
    def _iter_fields(self):
        """
        Iterates over the set fields of the builder.
        
        This method is an iterable generator.
        
        Yields
        ------
        conversion : ``Conversion``
        value : `object`
        """
        return
        yield
    

    def serialise(self, serialization_configuration):
        """
        Serialises the builder.
        
        Parameters
        ----------
        serialization_configuration : ``SerializationConfiguration``
            Configuration for serialization.
        
        Returns
        -------
        data : `dict<str, object> | FormData>`
        """
        data = {}
        defaults = serialization_configuration.defaults
        for conversion in serialization_configuration.conversions:
            for value in self._try_pull_field_value(conversion):
                data = conversion.serializer_putter(data, defaults, value)
        
        return data



from .integration import *
from .integration_account import *
from .integration_application import *
from .integration_metadata import *


__all__ = (
    *integration.__all__,
    *integration_account.__all__,
    *integration_application.__all__,
    *integration_metadata.__all__,
)



__all__ = ()

from ...field_parsers import entity_id_parser_factory, force_string_parser_factory
from ...field_putters import entity_id_putter_factory, force_string_putter_factory
from ...field_validators import entity_id_validator_factory, force_string_validator_factory

from .constants import NAME_LENGTH_MAX, NAME_LENGTH_MIN

# id

parse_id = entity_id_parser_factory('id')
put_id = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('guild_id')

# name

parse_name = force_string_parser_factory('name')
put_name = force_string_putter_factory('name')
validate_name = force_string_validator_factory('name', NAME_LENGTH_MIN, NAME_LENGTH_MAX)



__all__ = ()

from ..place_holder import PlaceHolder

from .slotted_meta import _get_direct_parent, _inherit_hash, _merge_type_slots, _process_set_slot, Slotted


ENTITY_ID_PLACEHOLDER = PlaceHolder(
    0,
    """
    Returns the discord entity's unique identifier number
    
    Returns
    -------
    id : `int`
    """
)


class DiscordEntityMeta(Slotted):
    """
    Meta type of Discord entities. Use ``DiscordEntity`` as parent type instead of using this type directly as a
    meta type.
    """
    def __new__(cls, type_name, type_parents, type_attributes, *, immortal = False):
        """
        Creates a Discord entity type. Sub-type ``DiscordEntity`` instead of using this type directly as a meta type.
        
        Parameters
        ----------
        type_name : `str`
            The created type's name.
        
        type_parents : `tuple<type>`
            The parent types of the creates type.
        
        type_attributes : `dict<str, object>`
            The type attributes of the created type.
        
        immortal : `bool` = `False`, Optional (Keyword only)
            Whether the created type's instances should support weakreferencing. If the inherited type supports
            weakreferencing, then the sub-type will as well of course.
        
        Returns
        -------
        type : ``DiscordEntityMeta``
        
        Notes
        -----
        The created instances are always slotted.
        
        When more types are inherited then use `__slots` at the secondary types for adding additional member
        descriptors.
        """
        direct_parent = _get_direct_parent(cls, type_name, type_parents)
        final_slots = _merge_type_slots(direct_parent, type_parents, type_attributes)
        _process_set_slot(type_attributes, final_slots)
        _inherit_hash(direct_parent, type_attributes)
        
        if immortal:
            for type_parent in type_parents:
                if hasattr(type_parent, '__weakref__'):
                    break
            else:
                final_slots.add('__weakref__')
        
        if (
            (direct_parent is not None) and
            ('id' not in type_attributes) and
            hasattr(direct_parent, 'id') and
            (direct_parent.id is ENTITY_ID_PLACEHOLDER)
        ):
            final_slots.add('id')
        
        type_attributes['__slots__'] = tuple(sorted(final_slots))
        
        return type.__new__(cls, type_name, type_parents, type_attributes)



__all__ = ('AuditLogEntryChangeConversionGroup', )

from scarletio import RichAttributeErrorBaseType


class AuditLogEntryChangeConversionGroup(RichAttributeErrorBaseType):
    """
    Represents how changes are converted of an audit log entry.
    
    Attributes
    ----------
    conversions : `tuple<AuditLogEntryChangeConversion>`
        The grouped conversions.
    key_pre_checker_conversions : `None | list<AuditLogEntryChangeConversion>`
        Conversions which have key pre checker.
    key_to_conversion : `None | dict<str, AuditLogEntryChangeConversion>`
        Field key to conversion relation used when deserializing.
    name_to_conversion : `None | dict<str, AuditLogEntryChangeConversion>`
        Name to conversion relation.
    """
    __slots__ = ('conversions', 'key_pre_checker_conversions', 'key_to_conversion', 'name_to_conversion',)
    
    def __new__(cls, *conversions):
        """
        Creates a new change conversion group with the given conversions..
        
        Parameters
        ----------
        *conversions : ``AuditLogEntryChangeConversion``
            Conversions to group.
        """
        key_pre_checker_conversions = None
        key_to_conversion = None
        name_to_conversion = None
        
        
        for conversion in conversions:
            if conversion.should_add_as_key_based_deserializer():
                for key in conversion.iter_field_keys():
                    if key_to_conversion is None:
                        key_to_conversion = {}
                    
                    key_to_conversion[key] = conversion
            
            if conversion.should_add_as_key_pre_check_deserializer():
                if key_pre_checker_conversions is None:
                    key_pre_checker_conversions = []
                
                key_pre_checker_conversions.append(conversion)
            
            if conversion.should_add_by_field_name():
                if name_to_conversion is None:
                    name_to_conversion = {}
                
                name_to_conversion[conversion.field_name] = conversion
        
        # Construct
        self = object.__new__(cls)
        self.conversions = conversions
        self.key_pre_checker_conversions = key_pre_checker_conversions
        self.key_to_conversion = key_to_conversion
        self.name_to_conversion = name_to_conversion
        return self
    
    
    def __repr__(self):
        """Returns the change conversion's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        # conversions
        repr_parts.append(' conversions = ')
        repr_parts.append(repr(self.conversions))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the entry change conversion's hash value."""
        return hash(self.conversions)
    
    
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        # conversions
        if self.conversions != other.conversions:
            return False
        
        return True
    
    
    def iter_field_keys(self):
        """
        Iterates over the field keys of the conversion group.
        
        This method is an iterable generator.
        
        Yields
        ------
        field_key : `str`
        """
        for conversion in self.conversions:
            yield from conversion.iter_field_keys()
    
    
    def get_conversion_for_key(self, key):
        """
        Gets the conversion for the given key.
        
        Parameters
        ----------
        key : `str`
            Conversion key.
        
        Returns
        -------
        conversion : `None | AuditLogEntryChangeConversion`
        """
        key_to_conversion = self.key_to_conversion
        if (key_to_conversion is not None):
            try:
                conversion = key_to_conversion[key]
            except KeyError:
                pass
            else:
                return conversion
        
        key_pre_checker_conversions = self.key_pre_checker_conversions
        if (key_pre_checker_conversions is not None):
            for conversion in key_pre_checker_conversions:
                if conversion.change_deserialization_key_pre_check(key):
                    return conversion
    
    
    def get_conversion_for_name(self, name):
        """
        Gets the conversion for the given name.
        
        Parameters
        ----------
        name : `str`
            Conversion name.
        
        Returns
        -------
        conversion : `None | AuditLogEntryChangeConversion`
        """
        name_to_conversion = self.name_to_conversion
        if (name_to_conversion is not None):
            return name_to_conversion.get(name, None)



__all__ = ()

from scarletio import Compound

from ...application import Application, Entitlement, SKU, Subscription
from ...application.entitlement.fields import (
    validate_exclude_deleted as validate_entitlement_exclude_deleted,
    validate_exclude_ended as validate_entitlement_exclude_ended, validate_guild_id as validate_entitlement_guild_id,
    validate_sku_ids as validate_entitlement_sku_ids, validate_user_id as validate_entitlement_user_id
)
from ...application.entitlement.utils import ENTITLEMENT_FIELD_CONVERTERS
from ...application.subscription.fields import (
    validate_sku_id as validate_subscription_sku_id, validate_user_id as validate_subscription_user_id
)
from ...embedded_activity import EmbeddedActivity
from ...http import DiscordApiClient
from ...payload_building import build_create_payload
from ...utils import log_time_converter

from ..request_helpers import (
    get_embedded_activity_and_id, get_entitlement_and_id, get_entitlement_id, get_subscription_and_sku_id_and_id
)

ENTITLEMENT_GET_CHUNK_LIMIT_MIN = 1
ENTITLEMENT_GET_CHUNK_LIMIT_MAX = 100


def _assert__application_id(application_id):
    """
    Asserts the the client's has `.application` synced by asserting whether it's value is not `0`.
    
    Parameters
    ----------
    application_id : `int`
        The client's application id.
    
    Raises
    ------
    AssertionError
        - If the client's application is not yet synced.
    """
    if application_id == 0:
        raise AssertionError(
            'The client\'s application is not yet synced.'
        )
    
    return True


class ClientCompoundApplicationEndpoints(Compound):
    
    api : DiscordApiClient
    application: Application
    
    
    async def entitlement_create(self, entitlement_template = None, **keyword_parameters):
        """
        Creates a new (test) entitlement.
        
        This method is a coroutine.
        
        Parameters
        ----------
        entitlement_template : `None`, ``Entitlement`` = `None`, Optional
            Entitlement to use as a template
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters either to define the template, or to overwrite specific fields' values.
        
        Other Parameters
        ----------------
        owner : ``None | ClientUserBase | Guild | (int | EntitlementOwnerType, int | str)``, Optional (Keyword only)
            The entitlement's owner. `owner_id` and `owner_type` combined.
        
        owner_id : `int`, Optional (Keyword only)
            The entitlement's owner's identifier.
        
        owner_type : ``EntitlementOwnerType``, `int`, Optional (Keyword only)
            The entitlement's owner's type.
        
        sku : ``int | SKU``, Optional (Keyword only)
            Alternative for `sku_id`.
        
        sku_id : ``int | SKU``, Optional (Keyword only)
            Stock keeping unit to create the entitlement for.
        
        Returns
        -------
        entitlement : ``Entitlement``
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        data = build_create_payload(entitlement_template, ENTITLEMENT_FIELD_CONVERTERS, keyword_parameters)
        entitlement_data = await self.api.entitlement_create(application_id, data)
        return Entitlement.from_data(entitlement_data)
    
    
    async def entitlement_delete(self, entitlement):
        """
        Deletes the given (test) entitlement.
        
        This method is a coroutine.
        
        Parameters
        ----------
        entitlement : ``Entitlement``, `int`
            The entitlement to delete.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        entitlement_id = get_entitlement_id(entitlement)
        await self.api.entitlement_delete(application_id, entitlement_id)
    
    
    async def entitlement_get_all(
        self, *, exclude_deleted = ..., exclude_ended = ..., guild_id = ..., sku_ids = ..., user_id = ...
    ):
        """
        Requests all the entitlements of the client's application's entitlements that satisfies the criteria.
        
        This method is a coroutine.
        
        Parameters
        ----------
        exclude_deleted : `bool`, Optional (Keyword only)
            Whether deleted entitlements should be excluded. Defaults to `False`.
        
        exclude_ended : `bool`, Optional (Keyword only)
            Whether ended entitlements should be excluded. Defaults to `False`.
        
        guild_id : `int`, `None`, ``Guild``, Optional (Keyword only)
            Guild identifier to filter for.
        
        sku_ids : ``None | int | SKU | iterable<int> | iterable<SKU>``, Optional (Keyword only)
            Stock keeping units to filter for.
        
        user_id : ``None | int | ClientUserBase``, Optional (Keyword only)
            User identifier to filter for.
        
        Returns
        -------
        entitlements : `list<Entitlement>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        query_string_parameters = {
            'limit': ENTITLEMENT_GET_CHUNK_LIMIT_MAX,
            'after': 0,
        }
        
        if user_id is not ...:
            user_id = validate_entitlement_user_id(user_id)
            if user_id:
                query_string_parameters['user_id'] = user_id
        
        if guild_id is not ...:
            guild_id = validate_entitlement_guild_id(guild_id)
            if guild_id:
                query_string_parameters['guild_id'] = guild_id
        
        if exclude_deleted is not ...:
            exclude_deleted = validate_entitlement_exclude_deleted(exclude_deleted)
            if exclude_deleted:
                query_string_parameters['exclude_deleted'] = exclude_deleted
        
        if exclude_ended is not ...:
            exclude_ended = validate_entitlement_exclude_ended(exclude_ended)
            if exclude_ended:
                query_string_parameters['exclude_ended'] = exclude_ended
        
        if sku_ids is not ...:
            sku_ids = validate_entitlement_sku_ids(sku_ids)
            if sku_ids is not None:
                query_string_parameters['sku_ids'] = ','.join(str(sku_id) for sku_id in sku_ids)
        
        entitlements = []
        
        while True:
            entitlement_datas = await self.api.entitlement_get_chunk(application_id, query_string_parameters)
            for entitlement_data in entitlement_datas:
                entitlement = Entitlement.from_data(entitlement_data)
                entitlements.append(entitlement)
                
            if len(entitlement_datas) < ENTITLEMENT_GET_CHUNK_LIMIT_MAX:
                break
            
            query_string_parameters['after'] = entitlements[-1].id
            continue
        
        return entitlements
    
    
    async def entitlement_get_chunk(
        self,
        *,
        after = ...,
        before = ...,
        exclude_deleted = ...,
        exclude_ended = ...,
        guild_id = ...,
        limit = ...,
        sku_ids = ...,
        user_id = ...,
    ):
        """
        Requests a chunk of the client's application's entitlements that satisfies the criteria.
        
        This method is a coroutine.
        
        Parameters
        ----------
        after : `int`, ``DiscordEntity``, `datetime`, Optional (Keyword only)
            The timestamp after the entitlements were created.
        
        before : `int`, ``DiscordEntity``, `datetime`, Optional (Keyword only)
            The timestamp before the entitlements were created.
        
        exclude_deleted : `bool`, Optional (Keyword only)
            Whether deleted entitlements should be excluded. Defaults to `False`.
        
        exclude_ended : `bool`, Optional (Keyword only)
            Whether ended entitlements should be excluded. Defaults to `False`.
        
        guild_id : `int`, `None`, ``Guild``, Optional (Keyword only)
            Guild identifier to filter for.
        
        limit : `int`, Optional (Keyword only)
            Up to how much entitlements should be requested. Can be in range `[1, 100]`. Defaults to `100`.
        
        sku_ids : ``None | int | SKU | iterable<int> | iterable<SKU>``, Optional (Keyword only)
            Stock keeping units to filter for.
        
        user_id : ``None | int | ClientUserBase``, Optional (Keyword only)
            User identifier to filter for.
        
        Returns
        -------
        entitlements : `list<Entitlement>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        if limit is ...:
            limit = ENTITLEMENT_GET_CHUNK_LIMIT_MAX
        
        elif isinstance(limit, int):
            if limit < ENTITLEMENT_GET_CHUNK_LIMIT_MIN:
                limit = ENTITLEMENT_GET_CHUNK_LIMIT_MIN
            elif limit > ENTITLEMENT_GET_CHUNK_LIMIT_MAX:
                limit = ENTITLEMENT_GET_CHUNK_LIMIT_MAX
        
        else:
            raise TypeError(
                f'`limit` can be `None | int`, got {type(limit).__name__}; {limit!r}.'
            )
        
        query_string_parameters = {
            'limit': limit,
        }
        
        if (after is not ...):
            query_string_parameters['after'] = log_time_converter(after)
        
        if (before is not ...):
            query_string_parameters['before'] = log_time_converter(before)
        
        if exclude_deleted is not ...:
            exclude_deleted = validate_entitlement_exclude_deleted(exclude_deleted)
            if exclude_deleted:
                query_string_parameters['exclude_deleted'] = exclude_deleted
        
        if exclude_ended is not ...:
            exclude_ended = validate_entitlement_exclude_ended(exclude_ended)
            if exclude_ended:
                query_string_parameters['exclude_ended'] = exclude_ended
        
        if guild_id is not ...:
            guild_id = validate_entitlement_guild_id(guild_id)
            if guild_id:
                query_string_parameters['guild_id'] = guild_id
        
        if sku_ids is not ...:
            sku_ids = validate_entitlement_sku_ids(sku_ids)
            if sku_ids is not None:
                query_string_parameters['sku_ids'] = ','.join(str(sku_id) for sku_id in sku_ids)
        
        if user_id is not ...:
            user_id = validate_entitlement_user_id(user_id)
            if user_id:
                query_string_parameters['user_id'] = user_id
        
        entitlement_datas = await self.api.entitlement_get_chunk(application_id, query_string_parameters)
        return [Entitlement.from_data(entitlement_data) for entitlement_data in entitlement_datas]
    
    
    async def entitlement_get(self, entitlement, *, force_update = False):
        """
        Requests the entitlement of the application.
        
        This method is a coroutine.
        
        Parameters
        ----------
        entitlement : ``Entitlement``, `int`
            The entitlement, or its identifier representing it.
        
        force_update : `bool` = `False`, Optional (Keyword only)
            Whether the entitlement should be requested even if it supposed to be up to date.
        
        Returns
        -------
        entitlement : ``Entitlement``
        
        Raises
        ------
        TypeError
            - If `entitlement` is given as incorrect type.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        entitlement, entitlement_id = get_entitlement_and_id(entitlement)
        
        if (not force_update) and (entitlement is not None) and (not entitlement.partial):
            return entitlement
        
        entitlement_data = await self.api.entitlement_get(application_id, entitlement_id)
        
        if (entitlement is None):
            entitlement = Entitlement.from_data(entitlement_data)
        else:
            entitlement._set_attributes(entitlement_data)
        
        return entitlement
    
    
    async def entitlement_consume(self, entitlement):
        """
        Marks the one-time usage entitlement as consumed.
        
        This method is a coroutine.
        
        Parameters
        ----------
        entitlement : ``Entitlement``, `int`
            The entitlement to consume.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        entitlement_id = get_entitlement_id(entitlement)
        await self.api.entitlement_consume(application_id, entitlement_id)
    
    
    async def subscription_get_sku(self, subscription, *, force_update = False):
        """
        Requests an sku bound subscription.
        
        This method is a coroutine.
        
        Parameters
        ----------
        subscription : ``Subscription``, `(int, int)`
            The subscription to request or a tuple of 2 snowflakes (sku_id & subscription_id) representing it.
        
        force_update : `bool` = `False`, Optional (Keyword only)
            Whether the emoji should be requested even if it supposed to be up to date.
        
        Returns
        -------
        subscription : ``Subscription``
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        subscription, sku_id, subscription_id = get_subscription_and_sku_id_and_id(subscription)
                
        if (not force_update) and (subscription is not None) and (not subscription.partial):
            return subscription
        
        subscription_data = await self.api.subscription_get_sku(sku_id, subscription_id)
        
        if (subscription is None):
            subscription = Subscription.from_data(subscription_data)
        else:
            subscription._set_attributes(subscription_data)
        
        return subscription
    
    
    async def subscription_get_all_sku_user(self, sku, user):
        """
        Requests all the sku bound subscriptions of the client's application's for the given user.
        
        This method is a coroutine.
        
        Parameters
        ----------
        sku : ``int | SKU``
            Stock keeping unit to get subscriptions for.
        
        user : `int`, ``ClientUserBase``
            User to get the subscriptions for.
        
        Returns
        -------
        subscriptions : `list<Subscription>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        sku_id = validate_subscription_sku_id(sku)
        user_id = validate_subscription_user_id(user)
        
        query_string_parameters = {
            'limit': 100,
            'after': 0,
            'user_id': user_id,
        }
        
        subscriptions = []
        
        while True:
            subscription_datas = await self.api.subscription_get_chunk_sku_user(sku_id, query_string_parameters)
            for subscription_data in subscription_datas:
                subscription = Subscription.from_data(subscription_data)
                subscriptions.append(subscription)
                
            if len(subscription_datas) < 100:
                break
            
            query_string_parameters['after'] = subscriptions[-1].id
            continue
        
        return subscriptions
    
    
    async def subscription_get_chunk_sku_user(
        self,
        sku,
        user,
        *,
        after = ...,
        before = ...,
        limit = ...,
    ):
        """
        Requests a chunk of the sku bound subscriptions of the client's application's for the given user.
        
        This method is a coroutine.
        
        Parameters
        ----------
        sku : ``int | SKU``
            Stock keeping unit to get subscriptions for.
        
        user : `int`, ``ClientUserBase``
            User to get the subscriptions for.
        
        after : `int`, ``DiscordEntity``, `datetime`, Optional (Keyword only)
            The timestamp after the subscriptions were created.
        
        before : `int`, ``DiscordEntity``, `datetime`, Optional (Keyword only)
            The timestamp before the subscriptions were created.
        
        limit : `int`, Optional (Keyword only)
            Up to how much subscriptions should be requested. Can be in range `[1, 100]`. Defaults to `100`.
        
        Returns
        -------
        subscriptions : `list<Subscription>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        sku_id = validate_subscription_sku_id(sku)
        user_id = validate_subscription_user_id(user)
        
        if limit is ...:
            limit = 100
        
        elif isinstance(limit, int):
            if limit < 1:
                limit = 1
            elif limit > 100:
                limit = 100
        
        else:
            raise TypeError(
                f'`limit` can be `None | int`, got {type(limit).__name__}; {limit!r}.'
            )
        
        query_string_parameters = {
            'limit': limit,
            'user_id': user_id,
        }
        
        if (after is not ...):
            query_string_parameters['after'] = log_time_converter(after)
        
        if (before is not ...):
            query_string_parameters['before'] = log_time_converter(before)
        
        subscription_datas = await self.api.subscription_get_chunk_sku_user(sku_id, query_string_parameters)
        return [Subscription.from_data(subscription_data) for subscription_data in subscription_datas]
    
    
    async def sku_get_all(self):
        """
        Requests the client's application's stock keeping unit.
        
        This method is a coroutine.
        
        Returns
        -------
        skus : `list` of ``SKU`
            The received stock keeping units.
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        sku_datas = await self.api.sku_get_all(application_id)
        return [SKU.from_data(sku_data) for sku_data in sku_datas]
    
    
    async def embedded_activity_get(self, embedded_activity, *, force_update = False):
        """
        Requests the embedded activity of the application.
        
        This method is a coroutine.
        
        Parameters
        ----------
        embedded_activity : ``EmbeddedActivity``, `(int, int)`
            The embedded activity, or its identifier representing it.
        
        force_update : `bool` = `False`, Optional (Keyword only)
            Whether the embedded_activity should be requested even if it supposed to be up to date.
        
        Returns
        -------
        embedded_activity : ``EmbeddedActivity``
        
        Raises
        ------
        TypeError
            - If `embedded_activity` is given as incorrect type.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        embedded_activity, embedded_activity_id = get_embedded_activity_and_id(embedded_activity)
        
        if (not force_update) and (embedded_activity is not None) and (not embedded_activity.partial):
            return embedded_activity
        
        embedded_activity_data = await self.api.embedded_activity_get(application_id, embedded_activity_id)
        
        if (embedded_activity is None):
            embedded_activity = EmbeddedActivity.from_data(embedded_activity_data)
        else:
            embedded_activity._set_attributes(embedded_activity_data, False, 0)
        
        return embedded_activity



__all__ = ()

from ...field_parsers import int_parser_factory, preinstanced_parser_factory
from ...field_putters import int_putter_factory, preinstanced_putter_factory
from ...field_validators import int_conditional_validator_factory, preinstanced_validator_factory

from .preinstanced import GuildActivityOverviewActivityLevel


# level

parse_level = preinstanced_parser_factory(
    'activity_level', GuildActivityOverviewActivityLevel, GuildActivityOverviewActivityLevel.none
)
put_level = preinstanced_putter_factory('activity_level')
validate_level = preinstanced_validator_factory('level', GuildActivityOverviewActivityLevel)


# score

parse_score = int_parser_factory('activity_score', 0)
put_score = int_putter_factory('activity_score')
validate_score = int_conditional_validator_factory(
    'score',
    0,
    lambda score : score >= 0,
    '>= 0',
)



__all__ = ()

from types import ModuleType


def try_set_attribute(module, attribute_name, attribute_value):
    """
    Tries to set to the module the given attributes.
    
    Parameters
    ----------
    module : ``ModuleType``
        The module type to set the attribute to.
    attribute_name : `str`
        The attribute's name to set.
    attribute_value : `object`
        The attribute value.
    """
    try:
        setattr(module, attribute_name, attribute_value)
    except AttributeError:
        pass


def create_module_from_spec(spec):
    """
    Creates a module from the given spec.
    
    Parameters
    ----------
    spec : ``PluginModuleSpecType``
        The module spec to create it's module for.
    
    Returns
    -------
    module : `ModuleType`
    """
    module = ModuleType(spec.name)
    
    try_set_attribute(module, '__name__', spec.name)
    try_set_attribute(module, '__loader__', spec.loader)
    try_set_attribute(module, '__package__', spec.parent)
    try_set_attribute(module, '__spec__', spec)
    try_set_attribute(module, '__path__', spec.submodule_search_locations)
    try_set_attribute(module, '__file__', spec.origin)
    try_set_attribute(module, '__cached__', spec.cached)
    
    return module



__all__ = ()

from scarletio import Compound

from ...bases import maybe_snowflake
from ...channel import Channel
from ...exceptions import DiscordException, ERROR_CODES
from ...guild import Guild
from ...http import DiscordApiClient
from ...invite import Invite
from ...invite.invite.fields import validate_code
from ...invite.invite.utils import INVITE_GUILD_FIELD_CONVERTERS
from ...payload_building import build_create_payload
from ...permission.permission import PERMISSION_MASK_CREATE_INSTANT_INVITE

from ..request_helpers import get_channel_id, get_guild_and_id, get_guild_id


def _iter_channels_in_preference_order(guild):
    """
    Iterates over a guild's channel in order of preference.
    
    Parameters
    ----------
    guild : ``Guild``
        The guild to iterate its channels of.
    
    Yields
    ------
    channel : ``Channel``
    """
    channel = guild.rules_channel
    if (channel is not None):
        yield channel 
    
    channel = guild.system_channel
    if (channel is not None):
        yield channel
    
    channel = guild.widget_channel
    if (channel is not None):
        yield channel
    
    for channel in guild.channel_list_flattened:
        if channel.is_in_group_invitable():
            yield channel


def _get_preferred_channel_for_invite(client, guild):
    """
    Gets the preferred channel to create invite for.
    
    Parameters
    ----------
    client : ``Client``
        The client who is getting the channel.
    guild : ``Guild``
        The guild to get the channel of.
    
    Returns
    -------
    channel : ``None | Channel``
    """
    for channel in _iter_channels_in_preference_order(guild):
        if channel.cached_permissions_for(client) & PERMISSION_MASK_CREATE_INSTANT_INVITE:
            return channel


class ClientCompoundInviteEndpoints(Compound):
    
    api : DiscordApiClient
    
    
    async def invite_get_vanity(self, guild):
        """
        Returns the vanity invite of the given guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``int | Guild``
            The guild, what's invite will be returned.
        
        Returns
        -------
        invite : `None`, ``Invite``
            The vanity invite of the `guild`, `None` if it has no vanity invite.
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild`` nor as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild, guild_id = get_guild_and_id(guild)
        
        invite_data_vanity = await self.api.invite_get_vanity(guild_id)
        invite_data = await self.api.invite_get(invite_data_vanity['code'], {'with_counts': True})
        invite_data['uses'] = invite_data_vanity.get('uses', None)
        return Invite.from_data(invite_data)
    
    
    async def invite_edit_vanity(self, guild, vanity_code, *, reason = None):
        """
        Edits the given guild's vanity invite's code.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``
            The guild, what's invite will be edited.
        vanity_code : `str`
            The new code of the guild's vanity invite.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the guild's audit logs.
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild`` nor as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            If `vanity_code` was not given as `str`.
        """
        guild_id = get_guild_id(guild)
        
        vanity_code = validate_code(vanity_code)
        
        await self.api.invite_edit_vanity(guild_id, {'code': vanity_code}, reason)
    
    
    async def invite_create(self, channel, invite_template = None, **keyword_parameters):
        """
        Creates an invite at the given channel with the given parameters.
        
        To create stream invite pass the `target_type` parameter as `InviteTargetTypes.stream` and use
        either the `target_user` or `target_user_id` to define the streamer.
        
        To create an embedded application invite pass the `target_type` parameter as
        `InviteTargetTypes.embedded_application` and use either the `target_application` or `target_application_id` to
        define the application.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The channel of the created invite.
        
        invite_template : `None`, ``Invite`` = `None`, Optional
            Invite entity to use as a template.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters to create the invite with.
        
        Other Parameters
        ----------------
        max_age : `int`, Optional (Keyword only)
            After how much time (in seconds) will the invite expire. Defaults is never.
        
        max_uses : `int`, Optional (Keyword only)
            How much times can the invite be used. Defaults to unlimited.
        
        target_application : `int`, ``Application``, Optional (Keyword only)
            Alternative for `target_application_id`.
        
        target_application_id : `int`, ``Application``, Optional (Keyword only)
            The invite's target application.
        
        target_type : ``None | int | InviteTargetType``, Optional (Keyword only)
            The invite's target type.
        
        target_user : `int`, ``ClientUserBase``, Optional (Keyword only)
            Alternative for `target_user`.
        
        target_user_id : `int`, ``ClientUserBase``, Optional (Keyword only)
            The target of the invite if applicable.
        
        temporary : `bool`, Optional (Keyword only)
            Whether the invite should give only temporary membership.
        
        unique : `bool`, Optional (Keyword only)
            Whether the created invite should be unique.
        
        Returns
        -------
        invite : ``Invite``
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id = get_channel_id(channel, Channel.is_in_group_invitable)
        data = build_create_payload(invite_template, INVITE_GUILD_FIELD_CONVERTERS, keyword_parameters)
        invite_data = await self.api.invite_create(channel_id, data)
        return Invite.from_data(invite_data)
    
    
    async def invite_create_preferred(self, guild, **keyword_parameters):
        """
        Creates an invite to the guild's preferred channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild . ``Guild``
            The guild the invite will be created to.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters to describe the created invite.
        
        Other Parameters
        ----------------
        max_age : `int` = `0`, Optional (Keyword only)
            After how much time (in seconds) will the invite expire. Defaults is never.
        
        max_uses : `int` = `0`, Optional (Keyword only)
            How much times can the invite be used. Defaults to unlimited.
        
        unique : `bool` = `True`, Optional (Keyword only)
            Whether the created invite should be unique.
        
        temporary : `bool` = `False`, Optional (Keyword only)
            Whether the invite should give only temporary membership.
        
        Returns
        -------
        invite : ``Invite``
        
        Raises
        ------
        TypeError
            - if `guild` is not ``Guild`˛
        ValueError
            If the guild has no channel to create invite to.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        if not isinstance(guild, Guild):
            raise TypeError(
                f'`guild` can be `{Guild.__name__}`, got {guild.__class__.__name__}; {guild!r}.'
            )
        
        channel = _get_preferred_channel_for_invite(self, guild)
        if channel is None:
            raise ValueError(
                f'The guild has no available channels to create invite for (or lack of permissions).'
                f'Got guild = {guild!r}.'
            )
        
        try:
            return (await self.invite_create(channel, **keyword_parameters))
        except DiscordException as err:
            if err.code in (
                ERROR_CODES.unknown_channel, # the channel was deleted meanwhile
                ERROR_CODES.missing_permissions, # permissions changed meanwhile
                ERROR_CODES.missing_access, # client removed
            ):
                return None
            raise
    
    
    async def invite_get(self, invite):
        """
        Requests a partial invite with the given code.
        
        This method is a coroutine.
        
        Parameters
        ----------
        invite : ``Invite``, `str`
            The invites code.
        
        Returns
        -------
        invite : ``Invite``
        
        Raises
        ------
        TypeError
            If `invite` was not given neither ``Invite`` nor `str`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            If `invite_code` was not given as `str`.
        """
        if isinstance(invite, Invite):
            invite_code = invite.code
        
        elif isinstance(invite, str):
            invite_code = invite
            invite = None
        
        else:
            raise TypeError(
                f'`invite`` can be `{Invite.__name__}`, `str`, got {type(invite).__name__}; '
                f'{invite!r}.'
            )
        
        invite_data = await self.api.invite_get(invite_code, {'with_counts': True, 'with_permissions': True})
        
        if invite is None:
            invite = Invite.from_data(invite_data)
        else:
            invite._update_attributes(invite_data)
        
        return invite
    
    
    async def invite_get_all_guild(self, guild):
        """
        Gets the invites of the given guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``int | Guild``
            The guild, what's invites will be requested.
        
        Returns
        -------
        invites : `list` of ``Invite`` objects
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild`` nor as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        
        invite_datas = await self.api.invite_get_all_guild(guild_id)
        return [Invite.from_data(invite_data) for invite_data in invite_datas]
    
    
    async def invite_get_all_channel(self, channel):
        """
        Gets the invites of the given channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The channel, what's invites will be requested.
        
        Returns
        -------
        invites : `list` of ``Invite`` objects
        
        Raises
        ------
        TypeError
            If `channel` was not given neither as ``Channel``, neither as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        while True:
            if isinstance(channel, Channel):
                if channel.is_in_group_invitable() or channel.partial:
                    channel_id = channel.id
                    break
            
            else:
                channel_id = maybe_snowflake(channel)
                if (channel_id is not None):
                    break
            
            raise TypeError(
                f'`channel` can be an invitable ``Channel``, `int`, got {channel.__class__.__name__}; {channel!r}.'
            )
        
        invite_datas = await self.api.invite_get_all_channel(channel_id)
        return [Invite.from_data(invite_data) for invite_data in invite_datas]
    
    
    async def invite_delete(self, invite, *, reason = None):
        """
        Deletes the given invite.
        
        This method is a coroutine.
        
        Parameters
        ----------
        invite : ``Invite``
            The invite to delete.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        Raises
        ------
        TypeError
            If `invite` was not given neither ``Invite`` nor `str`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        if isinstance(invite, Invite):
            invite_code = invite.code
        
        elif isinstance(invite, str):
            invite_code = invite
            invite = None
        
        else:
            raise TypeError(
                f'`invite`` can be `{Invite.__name__}`, `str`, got {invite.__class__.__name__}; {invite!r}.'
            )
        
        invite_data = await self.api.invite_delete(invite_code, reason)
        
        if invite is None:
            invite = Invite.from_data(invite_data)
        else:
            invite._update_attributes(invite_data)
        
        return invite
